# import core packages
import struct, copy, zlib, io, typing, logging, threading, shutil
from concurrent.futures import ThreadPoolExecutor
from tifffile import tifffile
from pathlib import Path

# Third party packages
import numpy
import imagecodecs

# bfio internals
import bfio
from bfio.OmeXml import OMEXML
from bfio import OmeXml
import bfio.base_classes

logging.basicConfig(format='%(asctime)s - %(name)-8s - %(levelname)-8s - %(message)s',
                    datefmt='%d-%b-%y %H:%M:%S')
logger = logging.getLogger('bfio.backends')

class PythonReader(bfio.base_classes.AbstractReader):

    logger = logging.getLogger('bfio.backends.PythonReader')

    _rdr = None

    def __init__(self, frontend):
        super().__init__(frontend)

        self.logger.debug('__init__(): Initializing _rdr (tifffile.TiffFile)...')
        self._rdr = tifffile.TiffFile(self.frontend._file_path)
        metadata = self.read_metadata()
        width = metadata.image().Pixels.get_SizeX()
        height = metadata.image().Pixels.get_SizeY()

        for tag in self._rdr.pages[0].tags:
            logger.debug(tag)

        tile_size = None

        if not self._rdr.pages[0].is_tiled or not self._rdr.pages[0].rowsperstrip:
            self.close()
            raise TypeError(frontend._file_path.name + ' is not a tiled tiff.' +
                            ' The python backend of the BioReader only ' +
                            'supports OME tiled tiffs. Use the java backend ' +
                            'to load this image.')

        elif self._rdr.pages[0].tilewidth != self.frontend._TILE_SIZE or \
            self._rdr.pages[0].tilelength != self.frontend._TILE_SIZE:

            if (width > frontend._TILE_SIZE or height > frontend._TILE_SIZE):
                self.close()
                raise ValueError('Tile width and height should be {} when '.format(self.frontend._TILE_SIZE) +
                                 'using the python backend, but found ' +
                                 'tilewidth={} and tilelength={}. Use the java '.format(self._rdr.pages[0].tilewidth,
                                                                                        self._rdr.pages[0].tilelength) +
                                 'backend to read this image.')

    def read_metadata(self):
        self.logger.debug('read_metadata(): Reading metadata...')
        return OMEXML(self._rdr.ome_metadata)

    def _chunk_indices(self,X,Y,Z):

        self.logger.debug('_chunk_indices(): (X,Y,Z) -> ({},{},{})'.format(X,Y,Z))
        assert len(X) == 2
        assert len(Y) == 2
        assert len(Z) == 2

        offsets = []
        bytecounts = []

        ts = self.frontend._TILE_SIZE

        x_tiles = numpy.arange(X[0]//ts,numpy.ceil(X[1]/ts),dtype=int)
        y_tile_stride = numpy.ceil(self.frontend.x/ts).astype(int)

        self.logger.debug('_chunk_indices(): x_tiles = {}'.format(x_tiles))
        self.logger.debug('_chunk_indices(): y_tile_stride = {}'.format(y_tile_stride))

        for z in range(Z[0],Z[1]):
            for y in range(Y[0]//ts,int(numpy.ceil(Y[1]/ts))):
                y_offset = int(y * y_tile_stride)
                ind = (x_tiles + y_offset).tolist()

                o = [self._rdr.pages[z].dataoffsets[i] for i in ind]
                b = [self._rdr.pages[z].databytecounts[i] for i in ind]

                self.logger.debug('_chunk_indices(): offsets = {}'.format(o))
                self.logger.debug('_chunk_indices(): bytecounts = {}'.format(b))

                offsets.extend(o)
                bytecounts.extend(b)

        return offsets,bytecounts

    def _process_chunk(self, args):

        keyframe = self._keyframe
        out = self._image

        w,l,d,_,_ = self._tile_indices[args[1]]

        # copy decoded segments to output array
        segment, _, shape = keyframe.decode(*args)

        if segment is None:
            segment = keyframe.nodata

        self.logger.debug('_process_chunk(): shape = {}'.format(shape))
        self.logger.debug('_process_chunk(): (w,l,d) = {},{},{}'.format(w[0],l[0],d[0]))

        out[l[0]: l[0] + shape[1],
            w[0]: w[0] + shape[2],
            d[0],0,0] = segment.squeeze()

    def _read_image(self,X,Y,Z,C,T,output):
        if (len(C)>1 and C[0]!=0) or (len(T)>0 and T[0]!=0):
            self.logger.warning('More than channel 0 was specified for either channel or timepoint data.' + \
                                'For the Python backend, only the first channel/timepoint will be loaded.')

        # Get keyframe
        self._keyframe = self._rdr.pages[0].keyframe
        fh = self._rdr.pages[0].parent.filehandle

        # Get binary data info
        offsets,bytecounts = self._chunk_indices(X,Y,Z)

        self.logger.debug('read_image(): _tile_indices = {}'.format(self._tile_indices))

        if self.frontend.max_workers > 1:
            with ThreadPoolExecutor(self.frontend.max_workers) as executor:
                executor.map(self._process_chunk,fh.read_segments(offsets,bytecounts))
        else:
            for args in fh.read_segments(offsets,bytecounts):
                self._process_chunk(args)

    def close(self):
        self._rdr.close()
        
    def __del__(self):
        self.close()

class PythonWriter(bfio.base_classes.AbstractWriter):
    _page_open = False
    _current_page = None

    logger = logging.getLogger('bfio.backends.PythonWriter')

    def __init__(self, frontend):
        super().__init__(frontend)

        if self.frontend.C > 1:
            self.logger.warning('The BioWriter only writes single channel ' +
                                'images, but the metadata has {} channels. '.format(self.frontend.C) +
                                'Setting the number of channels to 1.')
            self.frontend.C = 1
        if self.frontend.T > 1:
            self.logger.warning('The BioWriter only writes single timepoint ' +
                                'images, but the metadata has {} timepoints. '.format(self.frontend.T) +
                                'Setting the number of timepoints to 1.')
            self.frontend.T = 1

    def _pack(self, fmt, *val):
        return struct.pack(self._byteorder + fmt, *val)

    def _addtag(self,code, dtype, count, value, writeonce=False):
        tags = self._tags

        # compute ifdentry & ifdvalue bytes from code, dtype, count, value
        # append (code, ifdentry, ifdvalue, writeonce) to tags list
        if not isinstance(code, int):
            code = tifffile.TIFF.TAGS[code]
        try:
            tifftype = tifffile.TIFF.DATA_DTYPES[dtype]
        except KeyError as exc:
            raise ValueError(f'unknown dtype {dtype}') from exc
        rawcount = count

        if dtype == 's':
            # strings; enforce 7-bit ASCII on unicode strings
            if code == 270:
                value = tifffile.bytestr(value, 'utf-8') + b'\0'
            else:
                value = tifffile.bytestr(value, 'ascii') + b'\0'
            count = rawcount = len(value)
            rawcount = value.find(b'\0\0')
            if rawcount < 0:
                rawcount = count
            else:
                rawcount += 1  # length of string without buffer
            value = (value,)
        elif isinstance(value, bytes):
            # packed binary data
            dtsize = struct.calcsize(dtype)
            if len(value) % dtsize:
                raise ValueError('invalid packed binary data')
            count = len(value) // dtsize
        if len(dtype) > 1:
            count *= int(dtype[:-1])
            dtype = dtype[-1]
        ifdentry = [self._pack('HH', code, tifftype),
                    self._pack(self._writer._offsetformat, rawcount)]
        ifdvalue = None
        if struct.calcsize(dtype) * count <= self._writer._offsetsize:
            # value(s) can be written directly
            if isinstance(value, bytes):
                ifdentry.append(self._pack(self._writer._valueformat, value))
            elif count == 1:
                if isinstance(value, (tuple, list, numpy.ndarray)):
                    value = value[0]
                ifdentry.append(self._pack(self._writer._valueformat, self._pack(dtype, value)))
            else:
                ifdentry.append(self._pack(self._writer._valueformat,
                                self._pack(str(count) + dtype, *value)))
        else:
            # use offset to value(s)
            ifdentry.append(self._pack(self._writer._offsetformat, 0))
            if isinstance(value, bytes):
                ifdvalue = value
            elif isinstance(value, numpy.ndarray):
                if value.size != count:
                    raise RuntimeError('value.size != count')
                if value.dtype.char != dtype:
                    raise RuntimeError('value.dtype.char != dtype')
                ifdvalue = value.tobytes()
            elif isinstance(value, (tuple, list)):
                ifdvalue = self._pack(str(count) + dtype, *value)
            else:
                ifdvalue = self._pack(dtype, value)
        tags.append((code, b''.join(ifdentry), ifdvalue, writeonce))

    def _init_writer(self):
        """_init_writer Initializes file writing.

        This method is called exactly once per object. Once it is
        called, all other methods of setting metadata will throw an
        error.

        """

        self._tags = []

        self.frontend._metadata.image().set_ID(Path(self.frontend._file_path).name)

        if self.frontend.X * self.frontend.Y * self.frontend.bpp > 2**31:
            big_tiff = True
        else:
            big_tiff = False

        self._writer = tifffile.TiffWriter(self.frontend._file_path,
                                           bigtiff=big_tiff,
                                           append=False)

        self._byteorder = self._writer._byteorder

        self._datashape = (1, 1, 1) + (self.frontend.Y, self.frontend.X) + (1,)

        self._datadtype = numpy.dtype(self.frontend.dtype).newbyteorder(self._byteorder)

        tagnoformat = self._writer._tagnoformat
        valueformat = self._writer._valueformat
        offsetformat = self._writer._offsetformat
        offsetsize = self._writer._offsetsize
        tagsize = self._writer._tagsize

        # self._compresstag = tifffile.TIFF.COMPRESSION.NONE
        self._compresstag = tifffile.TIFF.COMPRESSION.ADOBE_DEFLATE

        # normalize data shape to 5D or 6D, depending on volume:
        #   (pages, planar_samples, height, width, contig_samples)
        self._samplesperpixel = 1
        self._bitspersample = self._datadtype.itemsize * 8

        self._tagbytecounts = 325  # TileByteCounts
        self._tagoffsets = 324  # TileOffsets

        def rational(arg, max_denominator=1000000):
            # return nominator and denominator from float or two integers
            from fractions import Fraction  # delayed import
            try:
                f = Fraction.from_float(arg)
            except TypeError:
                f = Fraction(arg[0], arg[1])
            f = f.limit_denominator(max_denominator)
            return f.numerator, f.denominator

        description = ''.join(['<?xml version="1.0" encoding="UTF-8"?>',
                                '<!-- Warning: this comment is an OME-XML metadata block, which contains crucial dimensional parameters and other important metadata. ',
                                'Please edit cautiously (if at all), and back up the original data before doing so. '
                                'For more information, see the OME-TIFF web site: https://docs.openmicroscopy.org/latest/ome-model/ome-tiff/. -->',
                                str(self.frontend._metadata).replace('ome:', '').replace(':ome', '')])

        self._addtag(270, 's', 0, description, writeonce=True)  # Description
        self._addtag(305, 's', 0, f'bfio v{bfio.__version__}')  # Software
        # addtag(306, 's', 0, datetime, writeonce=True)
        self._addtag(259, 'H', 1, self._compresstag)  # Compression
        self._addtag(256, 'I', 1, self._datashape[-2])  # ImageWidth
        self._addtag(257, 'I', 1, self._datashape[-3])  # ImageLength
        self._addtag(322, 'I', 1, self.frontend._TILE_SIZE)  # TileWidth
        self._addtag(323, 'I', 1, self.frontend._TILE_SIZE)  # TileLength

        sampleformat = {'u': 1, 'i': 2, 'f': 3, 'c': 6}[self._datadtype.kind]
        self._addtag(339, 'H', self._samplesperpixel,
                        (sampleformat,) * self._samplesperpixel)

        self._addtag(277, 'H', 1, self._samplesperpixel)
        self._addtag(258, 'H', 1, self._bitspersample)

        subsampling = None
        maxsampling = 1
        # PhotometricInterpretation
        self._addtag(262, 'H', 1, tifffile.TIFF.PHOTOMETRIC.MINISBLACK.value)

        if self.frontend.physical_size_x[0] is not None:
            self._addtag(282, '2I', 1,
                            rational(10000 / self.frontend.physical_size_x[0] ))  # XResolution in pixels/cm
            self._addtag(283, '2I', 1, rational(10000 / self.frontend.physical_size_y[0]))  # YResolution in pixels/cm
            self._addtag(296, 'H', 1, 3)  # ResolutionUnit = cm
        else:
            self._addtag(282, '2I', 1, (1, 1))  # XResolution
            self._addtag(283, '2I', 1, (1, 1))  # YResolution
            self._addtag(296, 'H', 1, 1)  # ResolutionUnit

        def bytecount_format(bytecounts, size=offsetsize):
            # return small bytecount format
            if len(bytecounts) == 1:
                return {4: 'I', 8: 'Q'}[size]
            bytecount = bytecounts[0] * 10
            if bytecount < 2 ** 16:
                return 'H'
            if bytecount < 2 ** 32:
                return 'I'
            if size == 4:
                return 'I'
            return 'Q'

        # can save data array contiguous
        contiguous = False

        # one chunk per tile per plane
        self._tiles = (
            (self._datashape[3] + self.frontend._TILE_SIZE - 1) // self.frontend._TILE_SIZE,
            (self._datashape[4] + self.frontend._TILE_SIZE - 1) // self.frontend._TILE_SIZE,
        )

        self._numtiles = tifffile.product(self._tiles)
        self._databytecounts = [
                                    self.frontend._TILE_SIZE ** 2 * self._datadtype.itemsize] * self._numtiles
        self._bytecountformat = bytecount_format(self._databytecounts)
        self._addtag(self._tagbytecounts, self._bytecountformat, self._numtiles, self._databytecounts)
        self._addtag(self._tagoffsets, self._writer._offsetformat, self._numtiles, [0] * self._numtiles)
        self._bytecountformat = self._bytecountformat * self._numtiles

        # the entries in an IFD must be sorted in ascending order by tag code
        self._tags = sorted(self._tags, key=lambda x: x[0])

    def _open_next_page(self):
        if self._current_page == None:
            self._current_page = 0
        else:
            self._current_page += 1

        if self._current_page == 1:
            for ind, tag in enumerate(self._tags):
                if tag[0] == 270:
                    del self._tags[ind]
                    break
            description = 'ImageJ=\nhyperstack=true\nimages={}\nchannels={}\nslices={}\nframes={}'.format(
                1, self.frontend.C, self.frontend.Z, self.frontend.T)
            self._addtag(270, 's', 0, description)  # Description
            self._tags = sorted(self._tags, key=lambda x: x[0])

        fh = self._writer._fh

        self._ifdpos = fh.tell()

        tagnoformat = self._writer._tagnoformat
        valueformat = self._writer._valueformat
        offsetformat = self._writer._offsetformat
        offsetsize = self._writer._offsetsize
        tagsize = self._writer._tagsize
        tagbytecounts = self._tagbytecounts
        tagoffsets = self._tagoffsets
        tags = self._tags

        if self._ifdpos % 2:
            # location of IFD must begin on a word boundary
            fh.write(b'\0')
            self._ifdpos += 1

        # update pointer at ifdoffset
        fh.seek(self._writer._ifdoffset)
        fh.write(self._pack(offsetformat, self._ifdpos))
        fh.seek(self._ifdpos)

        # create IFD in memory, do not write to disk
        if self._current_page < 2:
            self._ifd = io.BytesIO()
            self._ifd.write(self._pack(tagnoformat, len(tags)))
            tagoffset = self._ifd.tell()
            self._ifd.write(b''.join(t[1] for t in tags))
            self._ifdoffset = self._ifd.tell()
            self._ifd.write(self._pack(offsetformat, 0))  # offset to next IFD
            # write tag values and patch offsets in ifdentries
            for tagindex, tag in enumerate(tags):
                offset = tagoffset + tagindex * tagsize + offsetsize + 4
                code = tag[0]
                value = tag[2]
                if value:
                    pos = self._ifd.tell()
                    if pos % 2:
                        # tag value is expected to begin on word boundary
                        self._ifd.write(b'\0')
                        pos += 1
                    self._ifd.seek(offset)
                    self._ifd.write(self._pack(offsetformat, self._ifdpos + pos))
                    self._ifd.seek(pos)
                    self._ifd.write(value)
                    if code == tagoffsets:
                        self._dataoffsetsoffset = offset, pos
                    elif code == tagbytecounts:
                        self._databytecountsoffset = offset, pos
                    elif code == 270 and value.endswith(b'\0\0\0\0'):
                        # image description buffer
                        self._descriptionoffset = self._ifdpos + pos
                        self._descriptionlenoffset = (
                                self._ifdpos + tagoffset + tagindex * tagsize + 4)
                elif code == tagoffsets:
                    self._dataoffsetsoffset = offset, None
                elif code == tagbytecounts:
                    self._databytecountsoffset = offset, None
            self._ifdsize = self._ifd.tell()
            if self._ifdsize % 2:
                self._ifd.write(b'\0')
                self._ifdsize += 1

        self._databytecounts = [0 for _ in self._databytecounts]
        self._databyteoffsets = [0 for _ in self._databytecounts]

        # move to file position where data writing will begin
        # will write the tags later when the tile offsets are known
        fh.seek(self._ifdsize, 1)

        # write image data
        self._dataoffset = fh.tell()
        skip = (16 - (self._dataoffset % 16)) % 16
        fh.seek(skip, 1)
        self._dataoffset += skip

        self._page_open = True

    def _close_page(self):

        offsetformat = self._writer._offsetformat
        bytecountformat = self._bytecountformat

        fh = self._writer._fh

        # update strip/tile offsets
        offset, pos = self._dataoffsetsoffset
        self._ifd.seek(offset)
        if pos != None:
            self._ifd.write(self._pack(offsetformat, self._ifdpos + pos))
            self._ifd.seek(pos)
            for size in self._databyteoffsets:
                self._ifd.write(self._pack(offsetformat, size))
        else:
            self._ifd.write(self._pack(offsetformat, self._dataoffset))

        # update strip/tile bytecounts
        offset, pos = self._databytecountsoffset
        self._ifd.seek(offset)
        if pos:
            self._ifd.write(self._pack(offsetformat, self._ifdpos + pos))
            self._ifd.seek(pos)
        self._ifd.write(self._pack(bytecountformat, *self._databytecounts))

        self._fhpos = fh.tell()
        fh.seek(self._ifdpos)
        fh.write(self._ifd.getvalue())
        fh.flush()
        fh.seek(self._fhpos)

        self._writer._ifdoffset = self._ifdpos + self._ifdoffset

        # remove tags that should be written only once
        if self._current_page == 0:
            self._tags = [tag for tag in self._tags if not tag[-1]]

        self._page_open = False

    def _write_tiles(self, data, X, Y):

        assert len(X) == 2 and len(Y) == 2

        if X[0] % self.frontend._TILE_SIZE != 0 or Y[0] % self.frontend._TILE_SIZE != 0:
            logger.warning('X or Y positions are not on tile boundary, tile may save incorrectly')

        fh = self._writer._fh

        x_tiles = list(range(X[0] // self.frontend._TILE_SIZE, 1 + (X[1] - 1) // self.frontend._TILE_SIZE))
        tiles = []
        for y in range(Y[0] // self.frontend._TILE_SIZE, 1 + (Y[1] - 1) // self.frontend._TILE_SIZE):
            tiles.extend([y * self._tiles[1] + x for x in x_tiles])

        tile_shape = ((Y[1] - Y[0] - 1 + self.frontend._TILE_SIZE) // self.frontend._TILE_SIZE,
                        (X[1] - X[0] - 1 + self.frontend._TILE_SIZE) // self.frontend._TILE_SIZE)

        data = data.reshape(1, 1, 1, data.shape[0], data.shape[1], 1)
        tileiter = tifffile.iter_tiles(data,
                                        (self.frontend._TILE_SIZE, self.frontend._TILE_SIZE), tile_shape)

        # define compress function
        compressor = tifffile.TIFF.COMPESSORS[self._compresstag]

        def compress(data, level=1):

            return imagecodecs.deflate_encode(data,level)

        offsetformat = self._writer._offsetformat
        tagnoformat = self._writer._tagnoformat


        tileiter = [copy.deepcopy(tile) for tile in tileiter]
        if self.frontend.max_workers > 1:

            with ThreadPoolExecutor(max_workers=self.frontend.max_workers) as executor:

                compressed_tiles = iter(executor.map(compress, tileiter))

            for tileindex in tiles:
                t = next(compressed_tiles)
                self._databyteoffsets[tileindex] = fh.tell()
                fh.write(t)
                self._databytecounts[tileindex] = len(t)
        else:
            for tileindex, tile in zip(tiles,tileiter):

                t = compress(tile)
                self._databyteoffsets[tileindex] = fh.tell()
                fh.write(t)
                self._databytecounts[tileindex] = len(t)

        return None

    def close(self):
        """close_image Close the image

        This function should be called when an image will no longer be written
        to. This allows for proper closing and organization of metadata.
        """
        if self._writer != None:
            if self._page_open:
                self._close_page()
            self._ifd.close()
            self._writer._fh.close()

    def _write_image(self,X,Y,Z,C,T,image):

        if self._current_page != None and Z[0] < self._current_page:
            raise ValueError('Cannot write z layers below the current open page. (current page={},Z[0]={})'.format(
                self._current_page, Z[0]))

        # Do the work
        for zi, z in zip(range(0, Z[1] - Z[0]), range(Z[0], Z[1])):
            while z != self._current_page:
                if self._page_open:
                    self._close_page()
                self._open_next_page()
            self._write_tiles(image[..., zi, 0, 0], X, Y)


    def __del__(self):
        self.close()
try:
    import jpype
    import jpype.imports
    from jpype.types import *

    class JavaReader(bfio.base_classes.AbstractReader):

        logger = logging.getLogger('bfio.backends.JavaReader')
        _chunk_size = 4096
        _rdr = None
        _classes_loaded = False

        def _load_java_classes(self):
            if not jpype.isJVMStarted():
                bfio.start()

            global ImageReader
            from loci.formats import ImageReader

            global ServiceFactory
            from loci.common.services import ServiceFactory

            global OMEXMLService
            from loci.formats.services import OMEXMLService

            global IMetadata
            from loci.formats.meta import IMetadata

            JavaReader._classes_loaded = True

        def __init__(self, frontend):
            if not JavaReader._classes_loaded:
                self._load_java_classes()

            super().__init__(frontend)

            factory = ServiceFactory()
            service = factory.getInstance(OMEXMLService)
            # service = None
            self.omexml = service.createOMEXMLMetadata()

            self._rdr = ImageReader()
            self._rdr.setOriginalMetadataPopulated(True)
            self._rdr.setMetadataStore(self.omexml)
            self._rdr.setId(JString(str(self.frontend._file_path.absolute())))

        def read_metadata(self, update=False):

            return OMEXML(str(self.omexml.dumpXML()))

        def _read_image(self,X,Y,Z,C,T,output):

            out = self._image

            for ti,t in enumerate(T):
                for zi,z in enumerate(range(Z[0],Z[1])):
                    for ci,c in enumerate(C):
                        
                        # TODO: This should be changed in the future
                        # See the comment below about properly handling
                        # interleaved channel data.
                        if self.frontend.spp > 1:
                            index = self._rdr.getIndex(z, 0, t)
                        else:
                            index = self._rdr.getIndex(z, c, t)
                            
                        x_max = min([X[1],self.frontend.X])
                        for x in range(X[0],x_max,self._chunk_size):
                            x_range = min([self._chunk_size,x_max-x])

                            y_max = min([Y[1],self.frontend.Y])
                            for y in range(Y[0],y_max,self._chunk_size):
                                y_range = min([self._chunk_size,y_max-y])

                                image = self._rdr.openBytes(index, x, y, x_range, y_range)
                                image = numpy.frombuffer(bytes(image),self.frontend.dtype)
                                
                                # TODO: This should be changed in the future
                                # This reloads all channels for a tile on each
                                # loop. Ideally, there would be some better
                                # logic here to only load the necessary channel
                                # information once.
                                if self.frontend.spp > 1:
                                    image = image.reshape(self.frontend.c,y_range,x_range)
                                    image = image[c,...].squeeze()
                                    print(image.shape)
                                    
                                print(z)
                                out[y: y+y_range,
                                    x: x+x_range,
                                    zi,
                                    ci,
                                    ti] = image.reshape(y_range,x_range)

        def close(self):
            if jpype.isJVMStarted() and self._rdr != None:
                self._rdr.close()
                
        def __del__(self):
            self.close()

    class JavaWriter(bfio.base_classes.AbstractWriter):

        logger = logging.getLogger('bfio.backends.JavaWriter')

        # For Bioformats, the first tile has to be written before any other tile
        first_tile = False

        _classes_loaded = False

        def _load_java_classes(self):
            if not jpype.isJVMStarted():
                bfio.start()

            global OMETiffWriter
            from loci.formats.out import OMETiffWriter

            global ServiceFactory
            from loci.common.services import ServiceFactory

            global OMEXMLService
            from loci.formats.services import OMEXMLService

            global IMetadata
            from loci.formats.meta import IMetadata

            global CompressionType
            from loci.formats.codec import CompressionType
            
            JavaWriter._classes_loaded = True

        def __init__(self, frontend):
            if not JavaWriter._classes_loaded:
                self._load_java_classes()

            super().__init__(frontend)

            # Test to see if the loci_tools.jar is present
            if bfio.JARS == None:
                raise FileNotFoundError('The loci_tools.jar could not be found.')

        def _init_writer(self):
            """_init_writer Initializes file writing.

            This method is called exactly once per object. Once it is
            called, all other methods of setting metadata will throw an
            error.

            """
            if self.frontend._file_path.exists():
                self.frontend._file_path.unlink()

            writer = OMETiffWriter()

            # Set big tiff flag if file will be larger than 2GB
            if self.frontend.X * self.frontend.Y * self.frontend.bpp > 2**31:
                writer.setBigTiff(True)
            else:
                writer.setBigTiff(False)

            # Set the metadata
            service = ServiceFactory().getInstance(OMEXMLService);
            xml = str(self.frontend.metadata).replace('<ome:', '<').replace('</ome:', '</')
            metadata = service.createOMEXMLMetadata(xml)
            writer.setMetadataRetrieve(metadata);

            # Set the file path
            writer.setId(str(self.frontend._file_path))

            # We only save a single channel, so interleaved is unecessary
            writer.setInterleaved(False)

            # Set compression to
            writer.setCompression(CompressionType.ZLIB.getCompression())

            # Set tile sizes
            x = writer.setTileSizeX(self.frontend._TILE_SIZE)
            y = writer.setTileSizeY(self.frontend._TILE_SIZE)

            self._writer = writer

        def _process_chunk(self, dims):

            out = self._image

            X,Y,Z,C,T = dims

            x_range = min([self.frontend.X, X[1]+1024]) - X[1]
            y_range = min([self.frontend.Y, Y[1]+1024]) - Y[1]

            self.logger.debug('_process_chunk(): dims = {}'.format(dims))
            pixel_buffer = out[Y[0]:Y[0]+y_range, X[0]:X[0]+x_range, Z[0], C[0], T[0]].tobytes()
            self._writer.saveBytes(Z[1], pixel_buffer, X[1], Y[1], x_range, y_range)

        def _write_image(self,X,Y,Z,C,T,image):

            if not self.first_tile:
                args = self._tile_indices.pop(0)
                if args[0][1] > self.frontend._TILE_SIZE or args[0][1] > self.frontend._TILE_SIZE:
                    raise ValueError('The first write using the java backend ' +
                                     'must include the first tile.')
                self._process_chunk(args)
                self.first_tile = True

            if self.frontend.max_workers > 1:
                with ThreadPoolExecutor(self.frontend.max_workers) as executor:
                    executor.map(self._process_chunk,self._tile_indices)
            else:
                for args in self._tile_indices:
                    self._process_chunk(args)

        def close(self):
            if jpype.isJVMStarted() and self._writer != None:
                self._writer.close()
                
        def __del__(self):
            
            self.close()

except ModuleNotFoundError:

    logger.warning('Java backend is not available. This could be due to a ' +
                   'missing dependency (jpype).')

    class JavaReader(bfio.base_classes.AbstractReader):

        def __init__(self, frontend):

            raise ImportError('JavaReader class unavailable. Could not import' +
                              ' jpype.')

    class JavaWriter(bfio.base_classes.AbstractWriter):

        def __init__(self, frontend):

            raise ImportError('JavaWriter class unavailable. Could not import' +
                              ' jpype.')

try:
    import zarr
    from numcodecs import Blosc, blosc
    # blosc.use_threads = True

    class ZarrReader(bfio.base_classes.AbstractReader):

        logger = logging.getLogger('bfio.backends.ZarrReader')

        def __init__(self,frontend):
            super().__init__(frontend)

            self.logger.debug('__init__(): Initializing _rdr (zarr)...')

            self._root = zarr.open(str(self.frontend._file_path.resolve()),
                                   mode='r')

            self._rdr = self._root['0']

        def read_metadata(self):
            self.logger.debug('read_metadata(): Reading metadata...')
            if 'metadata' in self._root.attrs.keys():
                return OMEXML(self._root.attrs['metadata'])
            else:
                # Couldn't find OMEXML metadata, scrape metadata from file
                omexml = OMEXML()
                omexml.image(0).Name = Path(self.frontend._file_path).name
                p = omexml.image(0).Pixels

                for i,d in enumerate('XYZCT'):
                    setattr(p,'Size{}'.format(d),self._rdr.shape[4-i])
                
                p.channel_count = p.SizeC
                for i in range(0, p.SizeC):
                    p.Channel(i).Name = ''

                p.DimensionOrder = OmeXml.DO_XYZCT
                
                dtype = numpy.dtype(self._rdr.dtype.name).type
                for k,v in self.frontend._DTYPE.items():
                    if dtype==v:
                        p.PixelType = k
                        break

                return omexml

        def _process_chunk(self, dims):

            X,Y,Z,C,T = dims

            ts = self.frontend._TILE_SIZE

            data = self._rdr[T[1],C[1],
                             Z[1]:Z[1]+1,
                             Y[1]:Y[1]+ts,
                             X[1]:X[1]+ts]

            self._image[Y[0]:Y[0]+data.shape[-2],
                        X[0]:X[0]+data.shape[-1],
                        Z[0]:Z[0]+1,C[0],T[0]] = data.transpose(1,2,0)

        def _read_image(self,X,Y,Z,C,T,output):

            if self.frontend.max_workers > 1:
                with ThreadPoolExecutor(self.frontend.max_workers) as executor:
                    executor.map(self._process_chunk,self._tile_indices)
            else:
                for args in self._tile_indices:
                    self._process_chunk(args)


        def close(self):
            pass

    class ZarrWriter(bfio.base_classes.AbstractWriter):

        logger = logging.getLogger('bfio.backends.ZarrWriter')

        def __init__(self,frontend):
            super().__init__(frontend)

        def _init_writer(self):
            """_init_writer Initializes file writing.

            This method is called exactly once per object. Once it is called,
            all other methods of setting metadata will throw an error.

            NOTE: For Zarr, it is not explicitly necessary to make the file
                  read-only once writing has begun. Thus functionality is mainly
                  incorporated to remain consistent with the OME TIFF formats.
                  In the future, it may be reasonable to not enforce read-only

            """
            if self.frontend._file_path.exists():
                shutil.rmtree(self.frontend._file_path)

            shape = (
                self.frontend.T,
                self.frontend.C,
                self.frontend.Z,
                self.frontend.Y,
                self.frontend.X
            )

            compressor = Blosc(cname='zstd', clevel=1,shuffle=Blosc.SHUFFLE)

            self._root = zarr.group(store=str(self.frontend._file_path.resolve()))

            self._root.attrs['metadata'] = str(self.frontend._metadata)
            
            self._root.attrs['multiscales'] = [{
                "version": "0.1",
                "name": self.frontend._file_path.name,
                "datasets": [{
                    "path": "0"
                }],
                "metadata": {
                    "method": "mean"
                }
            }]

            writer = self._root.zeros('0',
                                      shape=shape,
                                      chunks=(1,1,1,self.frontend._TILE_SIZE,self.frontend._TILE_SIZE),
                                      dtype=self.frontend.dtype,
                                      compressor=compressor)

            # This is recommended to do for cloud storage to increase read/write
            # speed, but it also increases write speed locally when threading.
            zarr.consolidate_metadata(str(self.frontend._file_path.resolve()))

            self._writer = writer

        def _process_chunk(self,dims):

            out = self._image

            X,Y,Z,C,T = dims

            y1e = min([Y[1]+self.frontend._TILE_SIZE,self.frontend._DIMS["Y"]])
            y0e = min([Y[0]+self.frontend._TILE_SIZE,out.shape[0]])
            x1e = min([X[1]+self.frontend._TILE_SIZE,self.frontend._DIMS["X"]])
            x0e = min([X[0]+self.frontend._TILE_SIZE,out.shape[1]])
            self._writer[T[1]:T[1]+1,C[1]:C[1]+1,Z[1]:Z[1]+1,Y[1]:y1e,X[1]:x1e] = \
                out[Y[0]:y0e,X[0]:x0e,Z[0]:Z[0]+1,C[0]:C[0]+1,T[0]:T[0]+1].transpose(4,3,2,0,1)

        def _write_image(self,X,Y,Z,C,T,image):

            if self.frontend.max_workers > 1:
                with ThreadPoolExecutor(self.frontend.max_workers) as executor:
                    executor.map(self._process_chunk,self._tile_indices)
            else:
                for args in self._tile_indices:
                    self._process_chunk(args)

        def close(self):
            pass

except ModuleNotFoundError:

    logger.warning('Zarr backend is not available. This could be due to a ' +
                   'missing dependency (i.e. zarr)')

    class ZarrReader(bfio.base_classes.AbstractReader):

        def __init__(self, frontend):

            raise ImportError('ZarrReader class unavailable. Could not import' +
                              ' zarr.')

    class ZarrWriter(bfio.base_classes.AbstractWriter):

        def __init__(self, frontend):

            raise ImportError('ZarrWriter class unavailable. Could not import' +
                              ' zarr.')



import abc, threading, logging
import numpy
from concurrent.futures import ThreadPoolExecutor
from queue import Queue
import multiprocessing, typing
from pathlib import Path
import bfio

class BioBase(object,metaclass=abc.ABCMeta) :
    """ Abstract class for reading/writing OME tiled tiff images
    
    Attributes:
        dtype: Gets/sets the pixel type (e.g. uint8)
        channel_names: Gets/sets the names of each channel
        samples_per_pixel: Numbers of numbers per pixel location
        bytes_per_pixel: Number of bytes per pixel
        x: Get/set number of pixels in the x-dimension (width)
        y: Get/set number of pixels in the y-dimension (height)
        z: Get/set number of pixels in the z-dimension (depth)
        c: Get/set number of channels in the image
        t: Get/set number of timepoints in the image
        physical_size_x: Get/set the physical size of the x-dimension
        physical_size_y: Get/set the physical size of the y-dimension
        physical_size_z: Get/set the physical size of the z-dimension
        metadata: OmeXml object for the image
        cnames: Same as channel_names
        spp: Same as samples_per_pixel
        bpp: Same as bytes_per_pixel
        X: Same as x attribute
        Y: Same as y attribute
        Z: Same as z attribute
        C: Same as c attribute
        T: same as t attribute
        ps_x: Same as physical_size_x
        ps_y: Same as physical_size_y
        ps_z: Same as physical_size_z

    """
    # Set constants for reading/writing images
    _MAX_BYTES = 2 ** 30
    _DTYPE = {'uint8': numpy.uint8,
              'int8': numpy.int8,
              'uint16': numpy.uint16,
              'int16': numpy.int16,
              'uint32': numpy.uint32,
              'int32': numpy.int32,
              'float': numpy.float32,
              'double': numpy.float64}
    _BPP = {'uint8': 1,
            'int8': 1,
            'uint16': 2,
            'int16': 2,
            'uint32': 4,
            'int32': 4,
            'float': 4,
            'double': 8}
    _TILE_SIZE = 2 ** 10
    _CHUNK_SIZE = None
    
    _DIMS = {}
    
    _READ_ONLY_MESSAGE = '{} is read-only.'
    
    # protected backend object for interfacing with the file on disk
    _backend = None
    
    # protected attribute to hold metadata
    _metadata = None
    
    # protected buffering variables for iterating over an image
    _raw_buffer = Queue(maxsize=1)  # only preload one supertile at a time
    _data_in_buffer = Queue(maxsize=1)
    _supertile_index = Queue()
    _pixel_buffer = None
    _fetch_thread = None
    _tile_thread = None
    _tile_x_offset = 0
    _tile_last_column = 0
    
    def __init__(self,
                 file_path: typing.Union[str,Path],
                 max_workers: typing.Optional[int] = None,
                 backend: typing.Optional[str] = None,
                 read_only: typing.Optional[bool] = True):
        """__init__ Initialize BioBase object

         Args:
            file_path: Path to output file
            max_workers: Number of threads to be used.
                Default is None.
            backend: Backend to use, must be 'python' or 'java'.
                Default is 'python'.
        """
        
        # Whether the object is read only
        self.__read_only = read_only
        
        # Internally, keep the file_path as a Path object
        if isinstance(file_path,str):
            file_path = Path(file_path)
        self._file_path = file_path
        
        # validate/set the backend
        if backend == None:
            extension = ''.join(self._file_path.suffixes)
            if extension.endswith('.ome.tif'):
                backend = 'python'
            elif extension.endswith('.ome.zarr'):
                backend = 'zarr'
            else:
                backend = 'java'
        elif backend.lower() not in ['python','java', 'zarr']:
            raise ValueError('Keyword argument backend must be one of ["python","java","zarr"]')
        self._backend_name = backend.lower()
        
        # Set the number of workers for multi-threaded loading
        if self._backend_name=='java' and max_workers != None:
            self.logger.warning('The max_workers keyword was present, but the java backend only operates with a single worker. Setting max_workers=1.')
            self.max_workers = 1
        else:
            self.max_workers = max_workers if max_workers != None else max([multiprocessing.cpu_count()//2,1])
        
        # Create an thread lock for the object
        self._lock =  threading.Lock()
    
    def __setitem__(self,keys: typing.Union[list,tuple],values: numpy.ndarray):
        raise NotImplementedError('Cannot set values for {} class.'.format(self.__class__.__name__))
    
    def __getitem__(self,keys):
        raise NotImplementedError('Cannot get values for {} class.'.format(self.__class__.__name__))
    
    def _parse_slice(self,keys):
        
        # Dimension ordering and index initialization
        dims = 'YXZCT'
        ind = {d:None for d in dims}
        
        # If an empty slice, load the whole image
        if not isinstance(keys,tuple):
            if isinstance(keys,slice) and keys.start == None and keys.stop == None and keys.step==None:
                pass
            else:
                raise ValueError('If only a single index is supplied, it must be an empty slice ([:]).')
            
        # If not an empty slice, parse the key tuple
        else:
            
            # At most, 5 indices can be indicated
            if len(keys) > 5:
                raise ValueError('Found {} indices, but at most 5 indices may be supplied.'.format(len(keys)))
            
            # If the first key is an ellipsis, read backwards
            if keys[0] == Ellipsis:
                
                # If the last key is an ellipsis, throw an error
                if keys[-1]==Ellipsis:
                    raise ValueError('Ellipsis (...) may be used in either the first or last index, not both.')
                
                dims = ''.join([d for d in reversed(dims)])
                keys = [k for k in reversed(keys)]
            
            # Get key values
            for dim,key in zip(dims,keys):
                
                if isinstance(key,slice):
                    start = 0 if key.start == None else key.start
                    stop = getattr(self,dim) if key.stop == None else key.stop
                    
                    # For CT dimensions, generate a list from slices
                    if dim in 'CT':
                        step = 1 if key.step is None else key.step
                        ind[dim] = list(range(start,stop,step))
                    
                    # For XYZ dimensions, get start and stop of slice, ignore step
                    else:
                        ind[dim] = [start,stop]
                        
                elif isinstance(key,(int,tuple,list)):
                    # Only the last three dimensions can use int, tuple, or list indexing
                    if dim in 'ZCT':
                        if isinstance(key,int):
                            ind[dim] = [key]
                            if dim == 'Z':
                                ind[dim].append(ind[dim][0]+1)
                        else:
                            ind[dim] = key
                    else:
                        raise ValueError('The index in position {} must be a slice type.'.format(dims.find(dim)))
                elif key==Ellipsis:
                    if dims.find(dim)+1 < len(keys):
                        raise ValueError('Ellipsis may only be used in the first or last index.')
                else:
                    raise ValueError('Did not recognize indexing value of type: {}'.format(type(key)))
                
        return ind
                
    @property
    def read_only(self) -> bool:
        """Returns true if object is ready only"""
        return self.__read_only

    @read_only.setter
    def read_only(self):
        raise AttributeError(self._READ_ONLY_MESSAGE.format('read_only'))
    
    def __getattribute__(self,name):
        # Get image dimensions using num_x, x, or X
        if (len(name)==1 and name.lower() in 'xyzct'):
            return getattr(self._metadata.image().Pixels,'get_Size{}'.format(name.upper()))()
        else:
            return object.__getattribute__(self,name)
        
    def __setattr__(self,name,args):
        # Set image dimensions, for example, using x or X
        if len(name)==1 and name.lower() in 'xyzct':
            self.__xyzct_setter(name,args)
        else:
            object.__setattr__(self,name,args)
    
    def __xyzct_setter(self,dimension,value):
        assert not self.__read_only, self._READ_ONLY_MESSAGE.format(dimension.lower())
        assert value >= 1, "{} must be >= 0".format(dimension.upper())
        setattr(self._metadata.image(0).Pixels,'Size{}'.format(dimension.upper()),value)
        self._DIMS[dimension.upper()] = value
        if dimension.upper() == 'C':
            self._metadata.image(0).Pixels.channel_count = value
        self._metadata.image().Pixels.tiffdata_count = self.Z * self.C * self.T
        
        count = 0
        for z in range(self.Z):
            for c in range(self.C):
                for t in range(self.T):
                    self._metadata.image().Pixels.tiffdata(count).FirstZ = z
                    self._metadata.image().Pixels.tiffdata(count).FirstC = c
                    self._metadata.image().Pixels.tiffdata(count).FirstT = t
                    self._metadata.image().Pixels.tiffdata(count).IFD = count
                    self._metadata.image().Pixels.tiffdata(count).plane_count = 1
                    count += 1
    
    """ ------------------------------ """
    """ -Get/Set Dimension Properties- """
    """ ------------------------------ """
    @property
    def channel_names(self) -> typing.List[str]:
        """Get the channel names for the image"""
        image = self._metadata.image()
        return [image.Pixels.Channel(i).Name for i in range(0, self.C)]
        
    @channel_names.setter
    def channel_names(self,cnames: typing.List[str]):
        assert not self.__read_only, self._READ_ONLY_MESSAGE.format("channel_names")
        assert len(cnames) == self.C, "Number of names does not match number of channels."
        for i in range(0, len(cnames)):
            self._metadata.image(0).Pixels.Channel(i).Name = '' if cnames[i] == None else cnames[i]

    @property
    def shape(self) -> typing.Tuple[int,int,int,int,int]:
        """The 5-dimensional shape of the image

        Returns:
            (:attr:`~.Y`, :attr:`~.X`, :attr:`~.Z`, :attr:`~.C`, :attr:`~.T`)
            shape of the image
        """
        return tuple(getattr(self,d) for d in 'yxzct')
        
    @shape.setter
    def shape(self,new_shape: typing.Tuple[int,int,int,int,int]):
        assert len(new_shape) == 5
        for s,d in zip(new_shape,'yxzct'):
            setattr(self,d,s)

    @property
    def cnames(self) -> typing.List[str]:
        """Same as :attr:`~.channel_names` """
        return self.channel_names
        
    @cnames.setter
    def cnames(self,cnames: typing.List[str]):
        assert not self.__read_only, self._READ_ONLY_MESSAGE.format("cnames")
        self.channel_names = cnames
            
    def __physical_size(self,dimension,psize,units):
        if psize != None and units != None:
            assert not self.__read_only, self._READ_ONLY_MESSAGE.format("physical_size_{}".format(dimension.lower()))
            setattr(self._metadata.image(0).Pixels,'PhysicalSize{}'.format(dimension.upper()),psize)
            setattr(self._metadata.image(0).Pixels,'PhysicalSize{}Unit'.format(dimension.upper()),units)

    @property
    def physical_size_x(self) -> typing.Tuple[float,str]:
        """Physical size of pixels in x-dimension

        Returns:
            Units per pixel, Units (i.e. "cm" or "mm")
        """
        return (self._metadata.image(0).Pixels.PhysicalSizeX, self._metadata.image(0).Pixels.PhysicalSizeXUnit)

    @physical_size_x.setter
    def physical_size_x(self,size_units: tuple):
        self.__physical_size('X',*size_units)
        
    @property
    def ps_x(self) -> typing.Tuple[float,str]:
        """Same as :attr:`~bfio.bfio.BioReader.physical_size_x`"""
        return self.physical_size_x

    @ps_x.setter
    def ps_x(self,size_units: tuple):
        self.__physical_size('X',*size_units)
        
    @property
    def physical_size_y(self) -> typing.Tuple[float,str]:
        """Physical size of pixels in y-dimension

        Returns:
            Units per pixel, Units (i.e. "cm" or "mm")
        """
        return (self._metadata.image(0).Pixels.PhysicalSizeY, self._metadata.image(0).Pixels.PhysicalSizeYUnit)

    @physical_size_y.setter
    def physical_size_y(self,size_units: tuple):
        self.__physical_size('Y',*size_units)
        
    @property
    def ps_y(self):
        """Same as :attr:`~bfio.bfio.BioReader.physical_size_y`"""
        return self.physical_size_y

    @ps_y.setter
    def ps_y(self,size_units: tuple):
        self.__physical_size('Y',*size_units)
        
    @property
    def physical_size_z(self) -> typing.Tuple[float,str]:
        """Physical size of pixels in z-dimension

        Returns:
            Units per pixel, Units (i.e. "cm" or "mm")
        """
        return (self._metadata.image(0).Pixels.PhysicalSizeZ, self._metadata.image(0).Pixels.PhysicalSizeZUnit)

    @physical_size_z.setter
    def physical_size_z(self,size_units: tuple):
        self.__physical_size('Z',*size_units)
        
    @property
    def ps_z(self):
        """Same as :attr:`~.physical_size_z`"""
        return self.physical_size_z

    @ps_z.setter
    def ps_z(self,size_units: tuple):
        self.__physical_size('Z',*size_units)

    """ -------------------- """
    """ -Validation methods- """
    """ -------------------- """
    
    def _val_xyz(self, xyz: typing.Union[int,list,tuple], axis: str) -> typing.List[int]:
        """_val_xyz Utility function for validating image dimensions

        Args:
            xyz: Pixel value of x, y, or z dimension.
                If None, returns the maximum range of the dimension
            axis: Must be 'x', 'y', or 'z'

        Returns:
            list of ints indicating the first and last index in the dimension
        """
        assert axis in 'XYZ'
        
        if xyz == None:
            xyz = [0,self._DIMS[axis]]
        else:
            if axis=='Z' and isinstance(xyz,int):
                xyz = [xyz,xyz+1]
            assert len(xyz) == 2, \
                '{} must be a list or tuple of length 2.'.format(axis)
            assert xyz[0] >= 0, \
                '{}[0] must be greater than or equal to 0.'.format(axis)
            assert xyz[1] <= self._DIMS[axis], \
                '{}[1] cannot be greater than the maximum of the dimension ({}).'.format(axis,self._DIMS[axis])
                
        return xyz

    def _val_ct(self, ct: typing.Union[int,list], axis: str) -> typing.List[int]:
        """_val_ct Utility function for validating image dimensions

        Args:
            ct: List of ints indicating the channels or timepoints to load
                If None, returns a list of ints
            axis: Must be 'c', 't'

        Returns:
            list of ints indicating the first and last index in the dimension
        """

        assert axis in 'CT'
        
        if ct == None:
            # number of timepoints
            ct = list(range(0, self._DIMS[axis]))
        else:
            assert numpy.any(numpy.greater(self._DIMS[axis], ct)), \
            'At least one of the {}-indices was larger than largest index ({}).'.format(axis, self._DIMS[axis] - 1)
            assert numpy.any(numpy.less_equal(0, ct)), \
            'At least one of the {}-indices was less than 0.'.format(axis)
            assert len(ct) != 0, \
            'At least one {}-index must be selected.'.format(axis)
            
        return ct

    """ ------------------- """
    """ -Pixel information- """
    """ ------------------- """
    
    @property
    def dtype(self) -> numpy.dtype:
        """The numpy pixel type of the data"""
        return self._DTYPE[self._metadata.image(0).Pixels.PixelType]
    
    @dtype.setter
    def dtype(self,dtype):
        assert not self.__read_only, self._READ_ONLY_MESSAGE.format("dtype")
        if dtype in [numpy.uint64, numpy.int64]:
            self.logger.warning(f"{dtype} is not supported by Bioformats, saving as numpy.float64.")
            dtype = numpy.float64
        assert dtype in self._DTYPE.values(), "Invalid data type."
        for k,v in self._DTYPE.items():
            if dtype==v:
                self._metadata.image(0).Pixels.PixelType = k
                return
        
    @property
    def samples_per_pixel(self) -> int:
        """Number of samples per pixel"""
        return self._metadata.image().Pixels.Channel().SamplesPerPixel
    
    @samples_per_pixel.setter
    def samples_per_pixel(self,
                          samples_per_pixel: int):
        self._metadata.image().Pixels.Channel().SamplesPerPixel = samples_per_pixel

    @property
    def spp(self):
        """Same as :attr:`.samples_per_pixel`"""
        return self.samples_per_pixel
    
    @spp.setter
    def spp(self,samples_per_pixel):
        self.samples_per_pixel(samples_per_pixel)
        
    @property
    def bytes_per_pixel(self) -> int:
        """Number of bytes per pixel"""
        return self._BPP[self._metadata.image().Pixels.get_PixelType()]
    
    @bytes_per_pixel.setter
    def bytes_per_pixel(self,
                        bytes_per_pixel: int):
        raise AttributeError('Bytes per pixel cannot be set. Change the dtype instead')
    
    @property
    def bpp(self):
        """Same as :attr:`.bytes_per_pixel`"""
        return self.bytes_per_pixel
    
    @bpp.setter
    def bpp(self,bytes_per_pixel: int):
        self.bytes_per_pixel = bytes_per_pixel

    """ -------------------------- """
    """ -Other Methods/Properties- """
    """ -------------------------- """
    @property
    def metadata(self) -> bfio.OmeXml.OMEXML:
        """Get the metadata for the image

        This function calls the Bioformats metadata parser, which extracts
        metadata from an image. This returns a reference to an OMEXML class,
        which is a convenient handler for the complex xml metadata created by
        Bioformats.

        Most basic metadata information have their own BioReader methods, such
        as image dimensions(i.e. x, y, etc). However, in some cases it may be
        necessary to access the underlying metadata class.
        
        Minor changes have been made to the original OMEXML class created for
        python-bioformats, so the original OMEXML documentation should assist
        those interested in directly accessing the metadata. In general, it is
        best to assign data using the object properties to ensure the metadata
        stays in sync with the file.

        For information on the OMEXML class:
        https://github.com/CellProfiler/python-bioformats/blob/master/bioformats/omexml.py

        Returns:
            OMEXML object for the image
        """        
        return self._metadata
    
    @metadata.setter
    def metadata(self,value):
        raise AttributeError('The metadata attribute is read-only. Components' +
                             ' of the metadata can be modified by getting' +
                             ' the metadata object and making changes, or by' +
                             ' changing the attriutes of the image.')

    def maximum_batch_size(self,
                           tile_size: typing.List[int],
                           tile_stride: typing.Union[typing.List[int],None] = None) -> int:
        """maximum_batch_size Maximum allowable batch size for tiling
        
        The pixel buffer only loads at most two supertiles at a time. If the
        batch size is too large, then the tiling function will attempt to create
        more tiles than what the buffer holds. To prevent the tiling function
        from doing this, there is a limit on the number of tiles that can be
        retrieved in a single call. This function determines what the largest
        number of retreivable batches is.
        
        Args:
            tile_size: The height and width of the tiles to retrieve
            tile_stride: If None, defaults to tile_size. *Defaults to None.*
        
        Returns:
            Maximum allowed number of batches that can be retrieved by the
                iterate method.
        """
        if tile_stride == None:
            tile_stride = tile_size

        xyoffset = [(tile_size[0] - tile_stride[0]) / 2, (tile_size[1] - tile_stride[1]) / 2]

        num_tile_rows = int(numpy.ceil(self.Y / tile_stride[0]))
        num_tile_cols = (1024 - xyoffset[1]) // tile_stride[1]
        if num_tile_cols == 0:
            num_tile_cols = 1

        return int(num_tile_cols * num_tile_rows)
    
    def close(self):
        """Close the image"""
        if self._backend is not None:
            self._backend.close()
        
    def __enter__(self):
        """Handle entrance to a context manager
        
        This code is called when a `with` statement is used. This allows a
        BioBase object to be used like this:
        
        with bfio.BioReader('Path/To/File.ome.tif') as reader:
            ...
            
        with bfio.BioWriter('Path/To/File.ome.tif') as writer:
            ...
        """
        return self
    
    def __del__(self):
        """Handle file deletion

        This code runs when an object is deleted..
        """
        self.close()
        
    
    def __exit__(self, type_class, value, traceback):
        """Handle exit from the context manager

        This code runs when exiting a `with` statement.
        """
        self.close()
        
class AbstractBackend(object,metaclass=abc.ABCMeta):
    
    @abc.abstractmethod
    def __init__(self,frontend):
        self.frontend = frontend
        self._lock = threading.Lock()

    def _image_io(self,X,Y,Z,C,T,image):
            
        # Define tile bounds
        ts = self.frontend._TILE_SIZE
        X_tile_shape = X[1] - X[0]
        Y_tile_shape = Y[1] - Y[0]
        Z_tile_shape = Z[1] - Z[0]
        
        # Set the output for asynchronous reading
        self._image = image

        # Set up the tile indices
        self._tile_indices = []
        for t in range(len(T)):
            for c in range(len(C)):
                for z in range(Z_tile_shape):
                    for y in range(0,Y_tile_shape,ts):
                        for x in range(0,X_tile_shape,ts):
                            self._tile_indices.append(((x,X[0]+x),
                                                       (y,Y[0]+y),
                                                       (z,Z[0]+z),
                                                       (c,C[c]),
                                                       (t,T[t])))
        
        self.logger.debug('_image_io(): _tile_indices = {}'.format(self._tile_indices))
        
    @abc.abstractmethod
    def close(self):
        pass

class AbstractReader(AbstractBackend):
    
    @abc.abstractmethod
    def __init__(self,frontend):
        super().__init__(frontend)
    
    @abc.abstractmethod
    def read_metadata(self):
        pass
    
    def read_image(self,*args):
        with self._lock:
            self._image_io(*args)
            self._read_image(*args)

    @abc.abstractmethod
    def _read_image(self,X,Y,Z,C,T,output):
        pass

class AbstractWriter(AbstractBackend):
    
    _writer = None
    
    @abc.abstractmethod
    def __init__(self,frontend):
        super().__init__(frontend)
        self.initialized = False
    
    def write_image(self,*args):
        with self._lock:
            if not self.initialized:
                self._init_writer()
                self.frontend.__read_only = True
                self.initialized = True
            
            self._image_io(*args)
            self._write_image(*args)
        
    @abc.abstractmethod
    def _init_writer(self):
        pass
    
    @abc.abstractmethod
    def _write_image(*args):
        pass


import unittest
import requests, io, pathlib, shutil, logging, sys
import bfio
import numpy as np

TEST_IMAGES = {
    '1884807.ome.zarr': 'https://s3.embassy.ebi.ac.uk/idr/zarr/v0.1/1884807.zarr/',
    'Plate1-Blue-A-12-Scene-3-P3-F2-03.czi': 'https://downloads.openmicroscopy.org/images/Zeiss-CZI/idr0011/Plate1-Blue-A_TS-Stinger/Plate1-Blue-A-12-Scene-3-P3-F2-03.czi',
    '0.tif': 'https://osf.io/j6aer/download',
    'img_r001_c001.ome.tif': 'https://github.com/usnistgov/WIPP/raw/master/data/PyramidBuilding/inputCollection/img_r001_c001.ome.tif'
}
    
TEST_DIR = pathlib.Path(__file__).with_name('data')

logging.basicConfig(format='%(asctime)s - %(name)-8s - %(levelname)-8s - %(message)s',
                    datefmt='%d-%b-%y %H:%M:%S')
logger = logging.getLogger("bfio.test")

if '-v' in sys.argv:
    logger.setLevel(logging.INFO)

# def setUpModule():
    
#     """ Download images for testing """
#     TEST_DIR.mkdir(exist_ok=True)
    
#     for file,url in TEST_IMAGES.items():
#         logger.info(f'setup - Downloading: {file}')
        
#         if not file.endswith('.ome.zarr'):
#             r = requests.get(url)
            
#             with open(TEST_DIR.joinpath(file),'wb') as fw:
                
#                 fw.write(r.content)
#         else:
            
#             base_path = TEST_DIR.joinpath(file)
#             base_path.mkdir()
#             base_path.joinpath('0').mkdir()
            
#             units = [
#                 '.zattrs',
#                 '.zgroup',
#                 '0/.zarray',
#                 '0/0.0.0.0.0',
#                 '0/0.1.0.0.0',
#                 '0/0.2.0.0.0',
#             ]
            
#             for u in units:
                
#                 with open(base_path.joinpath(u),'wb') as fw:
                    
#                     fw.write(requests.get(url+u).content)
            
# def tearDownModule():
#     """ Remove test images """
    
#     logger.info('teardown - Removing test images...')
#     shutil.rmtree(TEST_DIR)

class TestSimpleRead(unittest.TestCase):
    
    @classmethod
    def tearDownClass(self):
        """ Load the czi image, and save as a npy file for further testing. """
        
        with bfio.BioReader(TEST_DIR.joinpath('Plate1-Blue-A-12-Scene-3-P3-F2-03.czi')) as br:
            
            np.save(TEST_DIR.joinpath('4d_array.npy'),br[:])
        
    def test_java(self):
        """test_java - Fails if Java/JPype improperly configured """
        
        bfio.start()
                
    def test_read_czi(self):
        """test_read_czi
        
        This test will fail if JPype and Java are not installed or improperly
        configured.
        
        """
        
        with bfio.BioReader(TEST_DIR.joinpath('Plate1-Blue-A-12-Scene-3-P3-F2-03.czi')) as br:
            
            self.assertEqual(br._backend_name, 'java')
            
            I = br[:]
            
            np.save(TEST_DIR.joinpath('4d_array.npy'),br[:])
            
            self.assertEqual(I.shape[0],512)
            self.assertEqual(I.shape[1],672)
            self.assertEqual(I.shape[2],21)
            self.assertEqual(I.shape[3],3)
            self.assertEqual(I.shape[4],1)
            self.assertEqual(I.dtype,np.uint16)
            
    def test_read_tif_strip_auto(self):
        """test_read_tif_strip_auto - Read tiff saved in strips, should load java backend """
        with bfio.BioReader(TEST_DIR.joinpath('0.tif')) as br:
            
            self.assertEqual(br._backend_name, 'java')
            
            I = br[:]
            
    def test_read_zarr_auto(self):
        """test_read_zarr_auto - Read ome zarr, should load zarr backend """
        with bfio.BioReader(TEST_DIR.joinpath('1884807.ome.zarr')) as br:
            
            self.assertEqual(br._backend_name, 'zarr')
            
            I = br[:]
            
            logger.info(I.shape)
            
    @unittest.expectedFailure
    def test_read_ome_tif_strip_auto(self):
        """test_read_ome_tif_strip_auto - Expected failure, should load python backend """
        with bfio.BioReader(TEST_DIR.joinpath('img_r001_c001.ome.tif')) as br:
            
            I = br[:]
            
    def test_read_tif_strip_java(self):
        """test_read_tif_strip_java - Read tiff using Java backend """
        with bfio.BioReader(TEST_DIR.joinpath('img_r001_c001.ome.tif'),backend='java') as br:
            
            self.assertEqual(br._backend_name,'java')
            
            I = br[:]
            
    @unittest.expectedFailure
    def test_read_tif_strip_python(self):
        """test_read_tif_strip_python - Expected failure, read tiff saved in strips """
        with bfio.BioReader(TEST_DIR.joinpath('img_r001_c001.ome.tif'),backend='python') as br:
            
            I = br[:]
            
""" Metadata tests to run on each backend """
def get_dims(reader):
    """ Get all dimension attributes """
    for dim in 'xyzct':
        logger.info('image.{} = {}'.format(dim,getattr(reader,dim)))
    for dim in 'xyzct'.upper():
        logger.info('image.{} = {}'.format(dim,getattr(reader,dim)))
    logger.info('image.shape = {}'.format(reader.shape))
    
def get_pixel_size(reader):
    """ Get all pixel size attributes """
    for dim in 'xyz':
        attribute = 'physical_size_{}'.format(dim)
        logger.info('image.physical_size_{} = {}'.format(dim,
                                                   getattr(reader,attribute)))
    for dim in 'xyz':
        attribute = 'ps_{}'.format(dim)
        logger.info('image.ps_{} = {}'.format(dim,
                                        getattr(reader,attribute)))
        
def get_pixel_info(reader):
    """ Get pixel information (type, samples per pixel, etc) """
    logger.info('image.samples_per_pixel={}'.format(reader.samples_per_pixel))
    logger.info('image.spp={}'.format(reader.spp))
    logger.info('image.bytes_per_pixel={}'.format(reader.bytes_per_pixel))
    logger.info('image.bpp={}'.format(reader.bpp))
    logger.info('image.dtype={}'.format(reader.dtype))
    
def get_channel_names(reader):
    """ Get channel names attribute """
    logger.info('image.channel_names={}'.format(reader.channel_names))
    logger.info('image.cnames={}'.format(reader.cnames))

""" Test classes (where the testing actually happens) """
class TestVersion(unittest.TestCase):
    
    def test_bfio_version(self):
        """ Ensure bfio version is properly loaded """
        logger.info('__version__ = {}'.format(bfio.__version__))
        assert bfio.__version__ != '0.0.0'
        
    def test_jar_version(self):
        """ Load loci-tools.jar and get version """
        logger.info('JAR_VERSION = {}'.format(bfio.JAR_VERSION))
        assert bfio.__version__ != None
        
class TestJavaReader(unittest.TestCase):
    
    def test_get_dims(self):
        """ Testing metadata dimension attributes """
        with bfio.BioReader(TEST_DIR.joinpath('0.tif')) as br:
            get_dims(br)
        
    def test_get_pixel_size(self):
        """ Testing metadata pixel sizes """
        with bfio.BioReader(TEST_DIR.joinpath('0.tif')) as br:
            get_pixel_size(br)
        
    def test_get_pixel_info(self):
        """ Testing metadata pixel information """
        with bfio.BioReader(TEST_DIR.joinpath('0.tif')) as br:
            get_pixel_info(br)
        
    def test_get_channel_names(self):
        """ Testing metadata channel names """
        with bfio.BioReader(TEST_DIR.joinpath('0.tif')) as br:
            get_channel_names(br)

class TestZarrReader(unittest.TestCase):
    
    def test_get_dims(self):
        """ Testing metadata dimension attributes """
        with bfio.BioReader(TEST_DIR.joinpath('1884807.ome.zarr')) as br:
            get_dims(br)
        
    def test_get_pixel_size(self):
        """ Testing metadata pixel sizes """
        with bfio.BioReader(TEST_DIR.joinpath('1884807.ome.zarr')) as br:
            get_pixel_size(br)
        
    def test_get_pixel_info(self):
        """ Testing metadata pixel information """
        with bfio.BioReader(TEST_DIR.joinpath('1884807.ome.zarr')) as br:
            get_pixel_info(br)
        
    def test_get_channel_names(self):
        """ Testing metadata channel names """
        with bfio.BioReader(TEST_DIR.joinpath('1884807.ome.zarr')) as br:
            get_channel_names(br)
            
class TestZarrMetadata(unittest.TestCase):
    
    def test_set_metadata(self):
        """ Testing metadata dimension attributes """
        cname = ['test']
        
        image = np.load(TEST_DIR.joinpath('4d_array.npy'))
        
        with bfio.BioWriter(TEST_DIR.joinpath('test_cname.ome.zarr')) as bw:
            bw.cnames = cname
            bw.ps_x = (100,'nm')
            bw.shape = image.shape
            bw[:] = image
        
        with bfio.BioReader(TEST_DIR.joinpath('test_cname.ome.zarr')) as br:
            logger.info(br.cnames)
            logger.info(br.ps_x)
            self.assertEqual(br.cnames[0],cname[0])


from __future__ import absolute_import, unicode_literals
import pathlib, logging

logging.basicConfig(format='%(asctime)s - %(name)-8s - %(levelname)-8s - %(message)s',
                    datefmt='%d-%b-%y %H:%M:%S')
logger = logging.getLogger("bfio.init")

log_level = logging.WARNING

try:
    with open(pathlib.Path(__file__).parent.joinpath("VERSION"),'r') as fh:
        __version__ = fh.read()
except:
    logger.info('Could not find VERSION. This is likely due to using a local/cloned version of bfio.')
    __version__ = "0.0.0"
logger.info('VERSION = {}'.format(__version__))

if pathlib.Path(__file__).parent.joinpath('jars').is_dir():

    try:
        import jpype

        _jars_dir = pathlib.Path(__file__).parent.joinpath('jars')

        JAR_VERSION = "Use bfio.start() to get JAR_VERSION."

        JARS = [str(_jars_dir.joinpath('loci_tools.jar').absolute())]

        LOG4J = str(_jars_dir.joinpath('log4j.properties').absolute())
        
    except ModuleNotFoundError:
        JAR_VERSION = None
        JARS = None
        LOG4J = None
        logger.info('jpype has not been installed. Can only use Python backend for reading/writing images.')

else:
    JAR_VERSION = None
    JARS = None
    LOG4J = None
    logger.info('The loci_tools.jar is not present. Can only use Python backend for reading/writing images.')

from .bfio import BioReader, BioWriter, start



from pathlib import Path

from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()
    
with open("VERSION",'r') as fh:
    version = fh.read()
    with open("./bfio/VERSION",'w') as fw:
        fw.write(version)

setup(
    name="bfio",
    version=version,
    author="Nick Schaub",
    author_email="nick.schaub@nih.gov",
    description="Simple reading and writing classes for tiled tiffs using Bioformats.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    project_urls={
        'Documentation': 'https://bfio.readthedocs.io/en/latest/',
        'Source': 'https://github.com/Nicholas-Schaub/polus-plugins/tree/master/utils/polus-bfio-util'
    },
    packages=find_packages(),
    package_data = {"bfio": ['jars/*.jar','jars/*.properties','VERSION']},
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires = ['tifffile<=2020.11.18',
                        'imagecodecs>=2021.2.26',
                        'numpy>=1.20.1'],
    extras_require = {
        'jpype': [
            'jpype1>=1.2.1'
        ],
        'zarr': [
            'zarr>=2.6.1'
        ],
        'all': [
            'jpype1>=1.2.1',
            'zarr>=2.6.1'
        ]
    }
)


import typing, numpy, struct, logging
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path
from bfio import backends, OmeXml, JARS, JAR_VERSION, LOG4J
from bfio.base_classes import BioBase
import numpy as np
import json

try:
    import jpype
    import jpype.imports
    from jpype.types import *

    def start() -> str:
        """Start the jvm

        This function starts the jvm and imports all the necessary Java classes
        to read images using the Bioformats toolbox.
        
        Return:
            The Bio-Formats JAR version.
        """
        
        global JAR_VERSION

        if jpype.isJVMStarted():
            from loci.formats import FormatTools
            
            JAR_VERSION = FormatTools.VERSION
            return JAR_VERSION
        
        logging.getLogger("bfio.start").info("Starting the jvm.")
        jpype.startJVM("-Dlog4j.configuration=file:{}".format(LOG4J),classpath=JARS)

        from loci.formats import FormatTools

        JAR_VERSION = FormatTools.VERSION

        logging.getLogger("bfio.start").info('loci_tools.jar version = {}'.format(JAR_VERSION))
        
        return JAR_VERSION

except ModuleNotFoundError:

    def start():
        raise ModuleNotFoundError("Error importing jpype or a loci_tools.jar class.")

class BioReader(BioBase):
    """Read supported image formats using Bioformats

    This class handles file reading of multiple formats. It can read files from
    any Bioformats supported file format, but is specially optimized for
    handling the OME tiled tiff format.
    
    There are three backends: ``java``, ``python``, and ``zarr``. The ``java``
    backend directly uses Bio-Formats for file reading, and can read any format
    that is supported by Bio-Formats. The ``python`` backend will only read
    images in OME Tiff format with tile tags set to 1024x1024, and is
    significantly faster than the "java" backend for reading these types of tiff
    files. The ``zarr`` backend will only read OME Zarr files.
    
    File reading and writing are multi-threaded by default, except for the
    ``java`` backend which does not currently support threading. Half of the
    available CPUs detected by multiprocessing.cpu_count() are used to read
    an image.

    For for information, visit the Bioformats page:
    https://www.openmicroscopy.org/bio-formats/

    Note:
        In order to use the ``java`` backend, jpype must be installed.
    """

    logger = logging.getLogger("bfio.bfio.BioReader")

    def __init__(self,
                 file_path: typing.Union[str,Path],
                 max_workers: typing.Union[int,None] = None,
                 backend: typing.Optional = None) ->  None:
        """
        Args:
            file_path: Path to file to read
            max_workers: Number of threads used to read and image. *Default is
                half the number of detected cores.*
            backend: Can be ``python``, ``java``, or ``zarr``. If None, then 
                BioReader will try to autodetect the proper backend.
                *Default is python.*
        """

        # Initialize BioBase
        super(BioReader, self).__init__(file_path,
                                        max_workers=max_workers,
                                        backend=backend)

        # Ensure backend is supported
        if self._backend_name == 'python':
            self._backend = backends.PythonReader(self)
        elif self._backend_name == 'java':
            self._backend = backends.JavaReader(self)
        elif self._backend_name == 'zarr':
            self._backend = backends.ZarrReader(self)
            
            if max_workers == 2:
                self.max_workers = 1
                self.logger.debug("Setting max_workers to 1, since max_workers==2 runs slower. To change back, set the object property.")
        else:
            raise ValueError('backend must be "python", "java", or "zarr"')

        # Preload the metadata
        self._metadata = self._backend.read_metadata()

        # Get dims to speed up validation checks
        self._DIMS = {
            'X': self.X,
            'Y': self.Y,
            'Z': self.Z,
            'C': self.C,
            'T': self.T
        }

    def __getitem__(self,keys: typing.Union[tuple,slice]) -> numpy.ndarray:
        """Image loading using numpy-like indexing

        This is an abbreviated method of accessing the :attr:`~.read` method,
        where a portion of the image will be loaded using numpy-like slicing
        syntax. Up to 5 dimensions can be designated depending on the number of
        available dimensions in the image array (Y, X, Z, C, T).

        Note:
            Not all methods of indexing can be used, and some indexing will lead
            to unexpected results. For example, logical indexing cannot be used,
            and step sizes in slice objects is ignored for the first three
            indices. This means and index such as ``[0:100:2,0:100:2,0,0,0]``
            will return a 100x100x1x1x1 numpy array.

        Parameters:
            keys: numpy-like slicing used to load a section of an image.

        Returns:
            A numpy.ndarray where trailing empty dimensions are removed.

        Example:

            .. code-block:: python

                import bfio

                # Initialize the bioreader
                br = bfio.BioReader('Path/To/File.ome.tif')

                # Load and copy a 100x100 array of pixels
                a = br[:100,:100,:1,0,0]

                # Slice steps sizes are ignored for the first 3 indices, so this
                # returns the same as above
                a = br[0:100:2,0:100:2,0:1,0,0]

                # The last two dimensions can receive a tuple or list as input
                # Load the first and third channel
                a = br[:100,100,0:1,(0,2),0]

                # If the file is 3d, load the first 10 z-slices
                b = br[...,:10,0,0]
        """

        ind = self._parse_slice(keys)

        return self.read(**ind)

    def read(self,
             X: typing.Union[list,tuple,None] = None,
             Y: typing.Union[list,tuple,None] = None,
             Z: typing.Union[list,tuple,int,None] = None,
             C: typing.Union[list,tuple,int,None] = None,
             T: typing.Union[list,tuple,int,None] = None) -> numpy.ndarray:
        """Read the image

        Read the all or part of the image. A n-dimmensional numpy.ndarray is
        returned such that all trailing empty dimensions will be removed.

        For example, if an image is read and it represents an xz plane, then the
        shape will be [1,m,n].

        Args:
            X: The (min,max) range of pixels to load along the x-axis (columns).
                If None, loads the full range. *Defaults to None.*
            Y: The (min,max) range of pixels to load along the y-axis (rows). If
                None, loads the full range. *Defaults to None.*
            Z: The (min,max) range of pixels to load along the z-axis (depth).
                Alternatively, an integer can be passed to select a single
                z-plane. If None, loads the full range. *Defaults to None.*
            C: Values indicating channel indices to load. If None, loads the
                full range. *Defaults to None.*
            T: Values indicating timepoints to load. If None, loads the full
                range. *Defaults to None.*

        Returns:
            A 5-dimensional numpy array.
        """
        # Validate inputs
        X = self._val_xyz(X, 'X')
        Y = self._val_xyz(Y, 'Y')
        Z = self._val_xyz(Z, 'Z')
        C = self._val_ct(C, 'C')
        T = self._val_ct(T, 'T')

        # Define tile bounds
        X_tile_start = (X[0]//self._TILE_SIZE) * self._TILE_SIZE
        Y_tile_start = (Y[0]//self._TILE_SIZE) * self._TILE_SIZE
        X_tile_end = numpy.ceil(X[1]/self._TILE_SIZE).astype(int) * self._TILE_SIZE
        Y_tile_end = numpy.ceil(Y[1]/self._TILE_SIZE).astype(int) * self._TILE_SIZE
        X_tile_shape = X_tile_end - X_tile_start
        Y_tile_shape = Y_tile_end - Y_tile_start
        Z_tile_shape = Z[1]-Z[0]

        # Initialize the output
        output = numpy.zeros([Y_tile_shape,
                              X_tile_shape,
                              Z_tile_shape,
                              len(C),
                              len(T)],
                             dtype=self.dtype)

        # Read the image
        self._backend.read_image([X_tile_start,X_tile_end],
                                 [Y_tile_start,Y_tile_end],
                                 Z,
                                 C,T,output)

        return output[Y[0]-Y_tile_start:Y[1]-Y_tile_start,X[0]-X_tile_start:X[1]-X_tile_start,...]

    def _fetch(self) -> numpy.ndarray:
        """Method for fetching image supertiles

        This method is intended to be run within a thread, and grabs a chunk of
        the image according to the coordinates in a queue.

        Currently, this function will only grab the first Z, C, and T positions
        regardless of what Z, C, and T coordinate are provided to the function.
        This function will need to be changed in then future to account for
        this.

        If the first value in X or Y is negative, then the image is pre-padded
        with the number of pixels equal to the absolute value of the negative
        number.

        If the last value in X or Y is larger than the size of the image, then
        the image is post-padded with the difference between the number and the
        size of the image. Input coordinate are read from the _supertile_index
        Queue object. Output data is stored in the _raw_buffer Queue object.

        As soon as the method is executed, a boolean value is put into the
        _data_in_buffer Queue to indicate that data is either in the buffer or
        will be put into the buffer.

        Returns:
            An image supertile
        """

        self._data_in_buffer.put(True)
        X, Y, Z, C, T = self._supertile_index.get()

        # Attach the jvm to the thread if present
        self._backend.attach()

        # Determine padding if needed
        reflect_x = False
        x_min = X[0]
        x_max = X[1]
        y_min = Y[0]
        y_max = Y[1]
        prepad_x = 0
        postpad_x = 0
        prepad_y = 0
        postpad_y = 0
        if x_min < 0:
            prepad_x = abs(x_min)
            x_min = 0
        if y_min < 0:
            prepad_y = abs(y_min)
            y_min = 0
        if x_max > self.num_x():
            if x_min >= self.num_x():
                x_min = self._TILE_SIZE * ((self.num_x() - 1) // self._TILE_SIZE)
                reflect_x = True
            x_max = self.num_x()
            postpad_x = x_max - self.num_x()
        if y_max > self.num_y():
            y_max = self.num_y()
            postpad_y = y_max - self.num_y()

        # Read the image
        I = self.read([x_min, x_max], [y_min, y_max], [0, 1], [0], [0]).squeeze()
        if reflect_x:
            I = numpy.fliplr(I)

        # Pad the image if needed
        if sum(1 for p in [prepad_x, prepad_y, postpad_x, postpad_y] if p != 0) > 0:
            I = numpy.pad(I, ((prepad_y, postpad_y), (prepad_x, postpad_x)), mode='symmetric')

        # Store the data in the buffer
        self._raw_buffer.put(I)

        # Detach the jvm
        self._backend.detach()

        return I

    def _buffer_supertile(self, column_start: int, column_end: int):
        """_buffer_supertile Process the pixel buffer

        Give the column indices of the data to process, and determine if the
        buffer needs to be processed. This method performs two operations on the
        buffer. First, it checks to see if data in the buffer can be shifted out
        of the buffer if it's already been processed, where data before
        column_start is assumed to have been processed. Second, this function
        loads data into the buffer if the image reader has made some available
        and there is room in _pixel_buffer for it.

        Args:
            column_start: First column index of data to be loaded
            column_end: Last column index of data to be loaded
        """

        # If the column indices are outside what is available in the buffer,
        # shift the buffer so more data can be loaded.
        if column_end - self._tile_x_offset >= self._TILE_SIZE:
            x_min = column_start - self._tile_x_offset
            x_max = self._pixel_buffer.shape[1] - x_min
            self._pixel_buffer[:, 0:x_max] = self._pixel_buffer[:, x_min:]
            self._pixel_buffer[:, x_max:] = 0
            self._tile_x_offset = column_start
            self._tile_last_column = numpy.argwhere((self._pixel_buffer == 0).all(axis=0))[0, 0]

        # Is there data in the buffer?
        if (self._supertile_index.qsize() > 0 or self._data_in_buffer.qsize() > 0):

            # If there is data in the _raw_buffer, return if there isn't room to load
            # it into the _pixel_buffer
            if self._pixel_buffer.shape[1] - self._tile_last_column < self._TILE_SIZE:
                return

            I = self._raw_buffer.get()
            if self._tile_last_column == 0:
                self._pixel_buffer[:I.shape[0], :I.shape[1]] = I
                self._tile_last_column = I.shape[1]
                self._tile_x_offset = column_start
            else:
                self._pixel_buffer[:I.shape[0], self._tile_last_column:self._tile_last_column + I.shape[1]] = I
                self._tile_last_column += I.shape[1]

            self._data_in_buffer.get()

    def _get_tiles(self,
                   X:typing.List[typing.List[int]],
                   Y:typing.List[typing.List[int]],
                   Z:typing.List[typing.List[int]],
                   C:typing.List[typing.List[int]],
                   T:typing.List[typing.List[int]]) -> numpy.ndarray:
        """_get_tiles Handle data buffering and tiling

        This function returns tiles of data according to the input coordinates.
        The X, Y, Z, C, and T are lists of lists, where each internal list
        indicates a set of coordinates specifying the range of pixel values to
        grab from an image.

        Args:
            X: List of 2-tuples indicating the (min,max) range of pixels to load
                within a tile.
            Y: List of 2-tuples indicating the (min,max) range of pixels to load
                within a tile.
            Z: Placeholder, to be implemented.
            C: Placeholder, to be implemented.
            T: Placeholder, to be implemented.

        Returns:
            numpy.ndarray: 2-dimensional ndarray.
        """

        self._buffer_supertile(X[0][0], X[0][1])

        if X[-1][0] - self._tile_x_offset > 1024:
            shift_buffer = True
            split_ind = 0
            while X[split_ind][0] - self._tile_x_offset < 1024:
                split_ind += 1
        else:
            shift_buffer = False
            split_ind = len(X)

        # Tile the data
        num_rows = Y[0][1] - Y[0][0]
        num_cols = X[0][1] - X[0][0]
        num_tiles = len(X)
        images = numpy.zeros((num_tiles, num_rows, num_cols, 1), dtype=self.pixel_type())

        for ind in range(split_ind):
            images[ind, :, :, 0] = self._pixel_buffer[Y[ind][0] - self._tile_y_offset:Y[ind][1] - self._tile_y_offset,
                                                      X[ind][0] - self._tile_x_offset:X[ind][1] - self._tile_x_offset]

        if split_ind != num_tiles:
            self._buffer_supertile(X[-1][0], X[-1][1])
            for ind in range(split_ind, num_tiles):
                images[ind, :, :, 0] = self._pixel_buffer[
                                       Y[ind][0] - self._tile_y_offset:Y[ind][1] - self._tile_y_offset,
                                       X[ind][0] - self._tile_x_offset:X[ind][1] - self._tile_x_offset]

        return images

    def __call__(self,
                 tile_size: typing.Union[list,tuple],
                 tile_stride: typing.Union[list,tuple,None] = None,
                 batch_size: typing.Union[int,None] = None,
                 channels: typing.Union[list] = [0]) -> typing.Iterable[typing.Tuple[numpy.ndarray,tuple]]:
        """Iterate through tiles of an image

        The BioReader object can be called, and will act as an iterator to load
        tiles of an image. The iterator buffers the loading of pixels
        asynchronously to quickly deliver images of the appropriate size.

        Args:
            tile_size: A list/tuple of length 2, indicating the height and width
                of the tiles to return.
            tile_stride: A list/tuple of length 2, indicating the row and column
                stride size. If None, then tile_stride = tile_size. *Defaults to
                None.*
            batch_size: Number of tiles to return on each iteration. *Defaults
                to None, which is the smaller of 32 or the*
                :attr:`~.maximum_batch_size`
            channels: A placeholder. Only the first channel
                is ever loaded. *Defaults to [0].*

        Returns:
            A tuple containing a 4-d numpy array and a tuple containing a list
            of X,Y,Z,C,T indices. The numpy array has dimensions
            ``[tile_num,tile_size[0],tile_size[1],channels]``

        Example:
            .. code:: python

                from bfio import BioReader
                import matplotlib.pyplot as plt

                br = BioReader('/path/to/file')

                for tiles,ind in br(tile_size=[256,256],tile_stride=[200,200]):
                    for i in tiles.shape[0]:
                        print('Displaying tile with X,Y coords: {},{}'.format(ind[i][0],ind[i][1]))
                        plt.figure()
                        plt.imshow(tiles[ind,:,:,0].squeeze())
                        plt.show()

        """

        self._iter_tile_size = tile_size
        self._iter_tile_stride = tile_stride
        self._iter_batch_size = batch_size
        self._iter_channels = channels

        return self

    def __iter__(self):

        tile_size = self._iter_tile_size
        tile_stride = self._iter_tile_stride
        batch_size = self._iter_batch_size
        channels = self._iter_channels

        if tile_size == None:
            raise SyntaxError('Cannot directly iterate over a BioReader object. Call it (i.e. for i in bioreader(256,256))')

        self._iter_tile_size = None
        self._iter_tile_stride = None
        self._iter_batch_size = None
        self._iter_channels = None

        # Ensure that the number of tiles does not exceed the supertile width
        if batch_size == None:
            batch_size = min([32, self.maximum_batch_size(tile_size, tile_stride)])
        else:
            assert batch_size <= self.maximum_batch_size(tile_size, tile_stride), \
                'batch_size must be less than or equal to {}.'.format(self.maximum_batch_size(tile_size, tile_stride))

        # input error checking
        assert len(tile_size) == 2, "tile_size must be a list with 2 elements"
        if tile_stride != None:
            assert len(tile_stride) == 2, "stride must be a list with 2 elements"
        else:
            stride = tile_size

        # calculate padding if needed
        if not (set(tile_size) & set(tile_stride)):
            xyoffset = [(tile_size[0] - tile_stride[0]) / 2, (tile_size[1] - tile_stride[1]) / 2]
            xypad = [(tile_size[0] - tile_stride[0]) / 2, (tile_size[1] - tile_stride[1]) / 2]
            xypad[0] = xyoffset[0] + (tile_stride[0] - numpy.mod(self.Y, tile_stride[0])) / 2
            xypad[1] = xyoffset[1] + (tile_stride[1] - numpy.mod(self.X, tile_stride[1])) / 2
            xypad = ((int(xyoffset[0]), int(2 * xypad[0] - xyoffset[0])),
                     (int(xyoffset[1]), int(2 * xypad[1] - xyoffset[1])))
        else:
            xyoffset = [0, 0]
            xypad = ((0, max([tile_size[0] - tile_stride[0], 0])),
                     (0, max([tile_size[1] - tile_stride[1], 0])))

        # determine supertile sizes
        y_tile_dim = int(numpy.ceil((self.Y - 1) / 1024))
        x_tile_dim = 1

        # Initialize the pixel buffer
        self._pixel_buffer = numpy.zeros((y_tile_dim * 1024 + tile_size[0], 2 * x_tile_dim * 1024 + tile_size[1]),
                                         dtype=self.dtype)
        self._tile_x_offset = -xypad[1][0]
        self._tile_y_offset = -xypad[0][0]

        # Generate the supertile loading order
        tiles = []
        y_tile_list = list(range(0, self.Y + xypad[0][1], 1024 * y_tile_dim))
        if y_tile_list[-1] != 1024 * y_tile_dim:
            y_tile_list.append(1024 * y_tile_dim)
        if y_tile_list[0] != xypad[0][0]:
            y_tile_list[0] = -xypad[0][0]
        x_tile_list = list(range(0, self.X + xypad[1][1], 1024 * x_tile_dim))
        if x_tile_list[-1] < self.X + xypad[1][1]:
            x_tile_list.append(x_tile_list[-1] + 1024)
        if x_tile_list[0] != xypad[1][0]:
            x_tile_list[0] = -xypad[1][0]
        for yi in range(len(y_tile_list) - 1):
            for xi in range(len(x_tile_list) - 1):
                y_range = [y_tile_list[yi], y_tile_list[yi + 1]]
                x_range = [x_tile_list[xi], x_tile_list[xi + 1]]
                tiles.append([x_range, y_range])
                self._supertile_index.put((x_range, y_range, [0, 1], [0], [0]))

        # Start the thread pool and start loading the first supertile
        thread_pool = ThreadPoolExecutor(self._max_workers)
        self._fetch_thread = thread_pool.submit(self._fetch)

        # generate the indices for each tile
        # TODO: modify this to grab more than just the first z-index
        X = []
        Y = []
        Z = []
        C = []
        T = []
        x_list = numpy.array(numpy.arange(-xypad[1][0], self.X, tile_stride[1]))
        y_list = numpy.array(numpy.arange(-xypad[0][0], self.Y, tile_stride[0]))
        for x in x_list:
            for y in y_list:
                X.append([x, x + tile_size[1]])
                Y.append([y, y + tile_size[0]])
                Z.append([0, 1])
                C.append(channels)
                T.append([0])

        # Set up batches
        batches = list(range(0, len(X), batch_size))

        # get the first batch
        b = min([batch_size, len(X)])
        index = (X[0:b], Y[0:b], Z[0:b], C[0:b], T[0:b])
        images = self._get_tiles(*index)

        # start looping through batches
        for bn in batches[1:]:
            # start the thread to get the next batch
            b = min([bn + batch_size, len(X)])
            self._tile_thread = thread_pool.submit(self._get_tiles, X[bn:b], Y[bn:b], Z[bn:b], C[bn:b], T[bn:b])

            # Load another supertile if possible
            if self._supertile_index.qsize() > 0 and not self._fetch_thread.running():
                self._fetch_thread = thread_pool.submit(self._fetch)

            # return the curent set of images
            yield images, index

            # get the images from the thread
            index = (X[bn:b], Y[bn:b], Z[bn:b], C[bn:b], T[bn:b])
            images = self._tile_thread.result()

        thread_pool.shutdown()

        # return the last set of images
        yield images, index

    @classmethod
    def image_size(cls,filepath: Path):
        """image_size Read image width and height from header
        
        This class method only reads the header information of tiff files or the
        zarr array json to identify the image width and height. There are
        instances when the image dimensions may want to be known without
        actually loading the image, and reading only the header is considerably
        faster than loading bioformats just to read simple metadata information.

        If the file is not a TIFF or OME Zarr, returns width = height = -1.

        This code was adapted to only operate on tiff images and includes
        additional to read the header of little endian encoded BigTIFF files.
        The original code can be found at:
        https://github.com/shibukawa/imagesize_py
        
        Args:
            filepath (pathlib.Path): Path to tiff file
        Returns:
            (width, height): Tuple of ints indicating width and height.

        """
        
        # Support strings input
        if isinstance(filepath,str):
            filepath = Path(filepath)
        
        # Handle a zarr file
        if filepath.name.endswith('ome.zarr'):
            with open(filepath.joinpath('0').joinpath('.zarray'),'r') as fr:
                zarray = json.load(fr)
                height = zarray['shape'][3]
                width = zarray['shape'][4]
            return width, height
        
        height = -1
        width = -1

        with open(str(filepath), 'rb') as fhandle:
            head = fhandle.read(24)
            size = len(head)

            # handle big endian TIFF
            if size >= 8 and head.startswith(b"\x4d\x4d\x00\x2a"):
                offset = struct.unpack('>L', head[4:8])[0]
                fhandle.seek(offset)
                ifdsize = struct.unpack(">H", fhandle.read(2))[0]
                for i in range(ifdsize):
                    tag, datatype, count, data = struct.unpack(">HHLL", fhandle.read(12))
                    if tag == 256:
                        if datatype == 3:
                            width = int(data / 65536)
                        elif datatype == 4:
                            width = data
                        else:
                            raise ValueError("Invalid TIFF file: width column data type should be SHORT/LONG.")
                    elif tag == 257:
                        if datatype == 3:
                            height = int(data / 65536)
                        elif datatype == 4:
                            height = data
                        else:
                            raise ValueError("Invalid TIFF file: height column data type should be SHORT/LONG.")
                    if width != -1 and height != -1:
                        break
                if width == -1 or height == -1:
                    raise ValueError("Invalid TIFF file: width and/or height IDS entries are missing.")
            # handle little endian Tiff
            elif size >= 8 and head.startswith(b"\x49\x49\x2a\x00"):
                offset = struct.unpack('<L', head[4:8])[0]
                fhandle.seek(offset)
                ifdsize = struct.unpack("<H", fhandle.read(2))[0]
                for i in range(ifdsize):
                    tag, datatype, count, data = struct.unpack("<HHLL", fhandle.read(12))
                    if tag == 256:
                        width = data
                    elif tag == 257:
                        height = data
                    if width != -1 and height != -1:
                        break
                if width == -1 or height == -1:
                    raise ValueError("Invalid TIFF file: width and/or height IDS entries are missing.")
            # handle little endian BigTiff
            elif size >= 8 and head.startswith(b"\x49\x49\x2b\x00"):
                bytesize_offset = struct.unpack('<L', head[4:8])[0]
                if bytesize_offset != 8:
                    raise ValueError('Invalid BigTIFF file: Expected offset to be 8, found {} instead.'.format(offset))
                offset = struct.unpack('<Q', head[8:16])[0]
                fhandle.seek(offset)
                ifdsize = struct.unpack("<Q", fhandle.read(8))[0]
                for i in range(ifdsize):
                    tag, datatype, count, data = struct.unpack("<HHQQ", fhandle.read(20))
                    if tag == 256:
                        width = data
                    elif tag == 257:
                        height = data
                    if width != -1 and height != -1:
                        break
                if width == -1 or height == -1:
                    raise ValueError("Invalid BigTIFF file: width and/or height IDS entries are missing.")

        return width, height

class BioWriter(BioBase):
    """BioWriter Write OME tiled tiff images

    This class handles the writing OME tiled tif images. There is a Java backend
    version of this tool that directly interacts with the Bioformats java
    library directly, and is primarily used for testing. It is currently not
    possible to change the tile size (which is set to 1024x1024).

    Unlike the BioReader class, the properties of this class are settable until
    the first time the ``write`` method is called.

    For for information, visit the Bioformats page:
    https://www.openmicroscopy.org/bio-formats/

    Note:
        In order to use the ``java`` backend, jpype must be installed.
    """

    logger = logging.getLogger("bfio.bfio.BioWriter")

    def __init__(self,
                 file_path: typing.Union[str,Path],
                 max_workers: typing.Union[int,None] = None,
                 backend: typing.Optional[str] = None,
                 metadata: typing.Union[OmeXml.OMEXML,None] = None,
                 image: typing.Union[numpy.ndarray,None] = None,
                 **kwargs) -> None:
        """

        Args:
            file_path: Path to file to read
            max_workers: Number of threads used to read and image. *Default is
                half the number of detected cores.*
            backend: Must be ``python`` or ``java``. *Default is python.*
            metadata: This directly sets the ome tiff metadata using the OMEXML
                class if specified. *Defaults to None.*
            image: The metadata will be set based on the dimensions and data
                type of the numpy array specified by this keyword argument.
                Ignored if metadata is specified. *Defaults to None.*
            kwargs: Most BioWriter object properties can be passed as keyword
                arguments to initialize the image metadata. If the metadata
                argument is used, then keyword arguments are ignored.
        """
        super(BioWriter, self).__init__(file_path=file_path,
                                        max_workers=max_workers,
                                        backend=backend,
                                        read_only=False)

        # Ensure backend is supported
        if self._backend_name == 'python':
            self._backend = backends.PythonWriter(self)
        elif self._backend_name == 'java':
            self._backend = backends.JavaWriter(self)
        elif self._backend_name == 'zarr':
            self._backend = backends.ZarrWriter(self)
        else:
            raise ValueError('backend must be "python", "java", or "zarr"')

        if metadata:
            assert metadata.__class__.__name__ == "OMEXML"
            self._metadata = OmeXml.OMEXML(str(metadata))
            self._metadata.image(0).Name = self._file_path.name
            self._metadata.image().Pixels.channel_count = self.C
            self._metadata.image().Pixels.DimensionOrder = OmeXml.DO_XYZCT
        else:
            self._metadata = self._minimal_xml()

            if isinstance(image, numpy.ndarray):
                assert len(image.shape) <= 5, "Image can be at most 5-dimensional (x,y,z,c,t)."
                self.spp = 1
                self.dtype = image.dtype
                for k,v in zip(image.shape,'YXZCT'):
                    setattr(self,k,v)

            elif kwargs:
                for k,v in kwargs.items():
                    setattr(self,k,v)

        if not self._file_path.name.endswith('.ome.tif') and not self._file_path.name.endswith('.ome.tif'):
            ValueError("The file extension must be .ome.tif or .ome.zarr")

        if self.metadata.image_count > 1:
            self.logger.warning('The BioWriter only writes single image ' +
                                'files, but the metadata has {} images. '.format(self.metadata.image_count) +
                                'Setting the number of images to 1.')
            self.metadata.image_count = 1

        # Get dims to speed up validation checks
        self._DIMS = {
            'X': self.X,
            'Y': self.Y,
            'Z': self.Z,
            'C': self.C,
            'T': self.T
        }

    def __setitem__(self,
                    keys: typing.Union[tuple,slice],
                    value: numpy.ndarray) -> None:
        """Image saving using numpy-like indexing

        This is an abbreviated method of accessing the :attr:`~.write` method,
        where a portion of the image will be saved using numpy-like slicing
        syntax. Up to 5 dimensions can be designated depending on the number of
        available dimensions in the image array (Y, X, Z, C, T).

        Note:
            Not all methods of indexing can be used, and some indexing will lead
            to unexpected results. For example, logical indexing cannot be used,
            and step sizes in slice objects is ignored for the first three
            indices. This means and index such as ``[0:100:2,0:100:2,0,0,0]``
            will save a 100x100x1x1x1 numpy array.

        Parameters:
            keys: numpy-like slicing used to save a section of an image.

        Example:

            .. code-block:: python

                import bfio

                # Initialize the biowriter
                bw = bfio.BioWriter('Path/To/File.ome.tif',
                                    X=100,
                                    Y=100,
                                    dtype=numpy.uint8)

                # Load and copy a 100x100 array of pixels
                bw[:100,:100,0,0,0] = np.zeros((100,100),dtype=numpy.uint8)

                # Slice steps sizes are ignored for the first 3 indices, so this
                # does the same as above
                bw[0:100:2,0:100:2] = np.zeros((100,100),dtype=numpy.uint8)

                # The last two dimensions can receive a tuple or list as input
                # Save two channels
                bw[:100,100,0,:2,0] = np.ones((100,100,1,2),dtype=numpy.uint8)

                # If the file is 3d, save the first 10 z-slices
                br[...,:10,0,0] = np.ones((100,100,1,2),dtype=numpy.uint8)
        """

        ind = self._parse_slice(keys)

        while len(value.shape) < 5:
            value = value[...,np.newaxis]

        for i,d in enumerate('YXZCT'):
            if ind[d] == None:
                if value.shape[i] != getattr(self,d):
                    raise IndexError('Shape of image {} does not match the '.format(value.shape) +
                                    'save dimensions {}.'.format((s[1] - s[0] for s in ind.values())))
            elif d in 'YXZ' and ind[d][1] - ind[d][0] != value.shape[i]:
                raise IndexError('Shape of image {} does not match the '.format(value.shape) +
                                 'save dimensions {}.'.format((s[1] - s[0] for s in ind.values())))
            elif d in 'CT' and len(ind[d]) != value.shape[i]:
                raise IndexError('Shape of image {} does not match the '.format(value.shape) +
                                 'save dimensions {}.'.format((s[1] - s[0] for s in ind.values())))
            elif d in 'YXZ':
                ind[d] = ind[d][0]

        self.write(value,**ind)

    def _minimal_xml(self) -> OmeXml.OMEXML:
        """Generates minimal xml for ome tif initialization

        Returns:
            OMEXML object
        """
        assert not self._backend._writer, "The image has started to be written. To modify the xml again, reinitialize."
        omexml = OmeXml.OMEXML()
        omexml.image(0).Name = Path(self._file_path).name
        p = omexml.image(0).Pixels

        assert isinstance(p, omexml.Pixels)

        for d in 'XYZCT':
            setattr(p,'Size{}'.format(d),1)

        p.DimensionOrder = OmeXml.DO_XYZCT
        p.PixelType = 'uint8'
        p.channel_count = 1

        return omexml

    def write(self,
              image: np.ndarray,
              X: typing.Union[int,None] = None,
              Y: typing.Union[int,None] = None,
              Z: typing.Union[int,None] = None,
              C: typing.Union[tuple,list,int,None] = None,
              T: typing.Union[tuple,list,int,None] = None) -> None:
        """write_image Write the image

        Write all or part of the image. A 5-dimmensional numpy.ndarray is
        required as the image input.

        Args:
            image: a 5-d numpy array
            X: The starting index of where to save data along the x-axis
                (columns). If None, loads the full range. *Defaults to None.*
            Y: The starting index of where to save data along the y-axis (rows).
                If None, loads the full range. *Defaults to None.*
            Z: The starting index of where to save data along the z-axis
                (depth). If None, loads the full range. *Defaults to None.*
            C: Values indicating channel indices to load. If None, loads the
                full range. *Defaults to None.*
            T: Values indicating timepoints to load. If None, loads the full
                range. *Defaults to None.*
        """
        # Set pixel bounds
        if X == None:
            X = 0
        if Y == None:
            Y = 0
        if Z == None:
            Z = 0

        if isinstance(X,int):
            X = [X]
        if isinstance(Y,int):
            Y = [Y]
        if isinstance(Z,int):
            Z = [Z]

        X.append(image.shape[1]+X[0])
        Y.append(image.shape[0]+Y[0])
        Z.append(image.shape[2]+Z[0])

        # Validate inputs
        X = self._val_xyz(X, 'X')
        Y = self._val_xyz(Y, 'Y')
        Z = self._val_xyz(Z, 'Z')
        C = self._val_ct(C, 'C')
        T = self._val_ct(T, 'T')

        assert len(image.shape) == 5, "Image must be 5-dimensional (x,y,z,c,t)."

        saving_shape = (Y[1]-Y[0],X[1]-X[0],Z[1]-Z[0],len(C),len(T))
        for d,v in zip(image.shape,saving_shape):
            if d != v:
                raise ValueError('Image shape {} does not match saving shape {}.'.format(image.shape,
                                                                                         saving_shape))

        # Define tile bounds
        X_tile_start = (X[0]//self._TILE_SIZE) * self._TILE_SIZE
        Y_tile_start = (Y[0]//self._TILE_SIZE) * self._TILE_SIZE
        X_tile_end = numpy.ceil(X[1]/self._TILE_SIZE).astype(int) * self._TILE_SIZE
        Y_tile_end = numpy.ceil(Y[1]/self._TILE_SIZE).astype(int) * self._TILE_SIZE
        X_tile_shape = X_tile_end - X_tile_start
        Y_tile_shape = Y_tile_end - Y_tile_start
        Z_tile_shape = Z[1]-Z[0]

        # if X[0] != X_tile_start or X[1] != X_tile_end or \
        #     Y[0] != Y_tile_start or Y[1] != Y_tile_end:
        #     self.logger.warning('Either X or Y dimension is not a multiple of the' +
        #                         ' tile size. This may result in the image being' +
        #                         ' saved incorrectly.')

        # Read the image
        self._backend.write_image([X_tile_start,X_tile_end],
                                  [Y_tile_start,Y_tile_end],
                                  Z,
                                  C,T,image)

    def close(self) -> None:
        """Close the image

        This function should be called when an image will no longer be written
        to. This allows for proper closing and organization of metadata.
        """
        if self._backend != None:
            self._backend.close()

    def _put(self):
        """_put Method for saving image supertiles

        This method is intended to be run within a thread, and writes a
        chunk of the image according to the coordinates in a queue.

        Currently, this function will only write the first Z, C, and T
        positions regardless of what Z, C, and T coordinate are provided
        to the function. This function will need to be changed in then
        future to account for this.

        If the last value in X or Y is larger than the size of the
        image, then the image is cropped to the appropriate size.

        Input coordinates are read from the _supertile_index Queue object.

        Input data is stored in the _raw_buffer Queue object.

        A boolean value is returned to indicate the processed has finished.
        """

        I = self._raw_buffer.get()
        X, Y, Z, C, T = self._supertile_index.get()

        # Attach the jvm to the thread
        self._backend.attach()

        # Write the image
        self.write(I[:self.num_y(), :, numpy.newaxis, numpy.newaxis, numpy.newaxis],
                   X=[X[0]],
                   Y=[Y[0]])

        # Detach the jvm
        self._backend.detach()

        return True

    def _buffer_supertile(self, column_start, column_end):
        """_buffer_supertile Process the pixel buffer

        Give the column indices of the data to process, and determine if
        the buffer needs to be processed. This method checks to see if
        data in the buffer can be shifted into the _raw_buffer for writing.

        Args:
            column_start ([int]): First column index of data to be loaded
            column_end ([int]): Last column index of data to be loaded

        """

        # If the start column index is outside of the width of the supertile,
        # write the data and shift the pixels
        if column_start - self._tile_x_offset >= 1024:
            self._raw_buffer.put(numpy.copy(self._pixel_buffer[:, 0:1024]))
            self._pixel_buffer[:, 0:1024] = self._pixel_buffer[:, 1024:2048]
            self._pixel_buffer[:, 1024:] = 0
            self._tile_x_offset += 1024
            self._tile_last_column = numpy.argwhere((self._pixel_buffer == 0).all(axis=0))[0, 0]

    def _assemble_tiles(self, images, X, Y, Z, C, T):
        """_assemble_tiles Handle data untiling

        This function puts tiles into the _pixel_buffer, effectively
        untiling them.

        Args:
            X (list): List of 2-tuples indicating the (min,max)
                range of pixels to load within a tile.
            Y (list): List of 2-tuples indicating the (min,max)
                range of pixels to load within a tile.
            Z (None): Placeholder, to be implemented.
            C (None): Placeholder, to be implemented.
            T (None): Placeholder, to be implemented.

        Returns:
            numpy.ndarray: 2-dimensional ndarray.
        """
        self._buffer_supertile(X[0][0], X[0][1])

        if X[-1][0] - self._tile_x_offset > 1024:
            split_ind = 0
            while X[split_ind][0] - self._tile_x_offset < 1024:
                split_ind += 1
        else:
            split_ind = len(X)

        # Untile the data
        num_rows = Y[0][1] - Y[0][0]
        num_cols = X[0][1] - X[0][0]
        num_tiles = len(X)

        for ind in range(split_ind):
            r_min = Y[ind][0] - self._tile_y_offset
            r_max = Y[ind][1] - self._tile_y_offset
            c_min = X[ind][0] - self._tile_x_offset
            c_max = X[ind][1] - self._tile_x_offset
            self._pixel_buffer[r_min:r_max, c_min:c_max] = images[ind, :, :, 0]

        if split_ind != num_tiles:
            self._buffer_supertile(X[-1][0], X[-1][1])
            for ind in range(split_ind, num_tiles):
                r_min = Y[ind][0] - self._tile_y_offset
                r_max = Y[ind][1] - self._tile_y_offset
                c_min = X[ind][0] - self._tile_x_offset
                c_max = X[ind][1] - self._tile_x_offset
                self._pixel_buffer[r_min:r_max, c_min:c_max] = images[ind, :, :, 0]

        self._tile_last_column = c_max

        return True

    def _writerate(self, tile_size, tile_stride=None, batch_size=None, channels=[0]):
        """writerate Image saving iterator

        This method is an iterator to save tiles of an image. This method
        buffers the saving of pixels asynchronously to quickly save
        images to disk. It is designed to work in complement to the
        BioReader.iterate method, and expects images to be fed into it in
        the exact same order as they would come out of that method.

        Data is sent to this iterator using the send() method once the
        iterator has been created. See the example for more information.

        Args:
            tile_size ([list,tuple]): A list/tuple of length 2, indicating
                the height and width of the tiles to return.
            tile_stride ([list,tuple], optional): A list/tuple of length 2,
                indicating the row and column stride size. If None, then
                tile_stride = tile_size.
                Defaults to None.
            batch_size (int, optional): Number of tiles to return on each
                iteration. Defaults to 32.
            channels (list, optional): A placeholder. Only the first channel
                is ever loaded. Defaults to [0].

        Yields:
            Nothing

        Example:
            from bfio import BioReader, BioWriter
            import numpy as np

            # Create the BioReader
            br = bfio.BioReader('/path/to/file')

            # Create the BioWriter
            out_path = '/path/to/output'
            bw = bfio.BioWriter(out_path,metadata=br.read_metadata())

            # Get the batch size
            batch_size = br.maximum_batch_size(tile_size=[256,256],tile_stride=[256,256])
            readerator = br.iterate(tile_size=[256,256],tile_stride=[256,256],batch_size=batch_size)
            writerator = bw.writerate(tile_size=[256,256],tile_stride=[256,256],batch_size=batch_size)

            # Initialize the writerator
            next(writerator)

            # Load tiles of the imgae and save them
            for images,indices in readerator:
                writerator.send(images)
            bw.close_image()

            # Verify images are the same
            original_image = br.read_image()
            bw = bfio.BioReader(out_path)
            saved_image = bw.read_image()

            print('Original and saved images are the same: {}'.format(numpy.array_equal(original_image,saved_image)))

        """

        # Enure that the number of tiles does not exceed the width of a supertile
        if batch_size == None:
            batch_size = min([32, self.maximum_batch_size(tile_size, tile_stride)])
        else:
            assert batch_size <= self.maximum_batch_size(tile_size, tile_stride), \
                'batch_size must be less than or equal to {}.'.format(
                    self.maximum_batch_size(tile_size, tile_stride))

        # input error checking
        assert len(tile_size) == 2, "tile_size must be a list with 2 elements"
        if tile_stride != None:
            assert len(tile_stride) == 2, "stride must be a list with 2 elements"
        else:
            stride = tile_size

        # calculate unpadding
        if not (set(tile_size) & set(tile_stride)):
            xyoffset = [int((tile_size[0] - tile_stride[0]) / 2), int((tile_size[1] - tile_stride[1]) / 2)]
            xypad = [(tile_size[0] - tile_stride[0]) / 2, (tile_size[1] - tile_stride[1]) / 2]
            xypad[0] = xyoffset[0] + (tile_stride[0] - numpy.mod(self.num_y(), tile_stride[0])) / 2
            xypad[1] = xyoffset[1] + (tile_stride[1] - numpy.mod(self.num_x(), tile_stride[1])) / 2
            xypad = ((int(xyoffset[0]), int(2 * xypad[0] - xyoffset[0])),
                        (int(xyoffset[1]), int(2 * xypad[1] - xyoffset[1])))
        else:
            xyoffset = [0, 0]
            xypad = ((0, max([tile_size[0] - tile_stride[0], 0])),
                        (0, max([tile_size[1] - tile_stride[1], 0])))

        # determine supertile sizes
        y_tile_dim = int(numpy.ceil((self.num_y() - 1) / 1024))
        x_tile_dim = 1

        # Initialize the pixel buffer
        self._pixel_buffer = numpy.zeros((y_tile_dim * 1024 + tile_size[0], 2 * x_tile_dim * 1024 + tile_size[1]),
                                        dtype=self.pixel_type())
        self._tile_x_offset = 0
        self._tile_y_offset = 0

        # Generate the supertile saving order
        tiles = []
        y_tile_list = list(range(0, self.num_y(), 1024 * y_tile_dim))
        if y_tile_list[-1] != 1024 * y_tile_dim:
            y_tile_list.append(1024 * y_tile_dim)
        x_tile_list = list(range(0, self.num_x(), 1024 * x_tile_dim))
        if x_tile_list[-1] < self.num_x() + xypad[1][1]:
            x_tile_list.append(x_tile_list[-1] + 1024)

        for yi in range(len(y_tile_list) - 1):
            for xi in range(len(x_tile_list) - 1):
                y_range = [y_tile_list[yi], y_tile_list[yi + 1]]
                x_range = [x_tile_list[xi], x_tile_list[xi + 1]]
                tiles.append([x_range, y_range])
                self._supertile_index.put((x_range, y_range, [0, 1], [0], [0]))

        # Start the thread pool and start loading the first supertile
        thread_pool = ThreadPoolExecutor(self._max_workers)

        # generate the indices for each tile
        # TODO: modify this to grab more than just the first z-index
        X = []
        Y = []
        Z = []
        C = []
        T = []
        x_list = numpy.array(numpy.arange(0, self.num_x(), tile_stride[1]))
        y_list = numpy.array(numpy.arange(0, self.num_y(), tile_stride[0]))
        for x in x_list:
            for y in y_list:
                X.append([x, x + tile_stride[1]])
                Y.append([y, y + tile_stride[0]])
                Z.append([0, 1])
                C.append(channels)
                T.append([0])

        # start looping through batches
        bn = 0
        while bn < len(X):
            # Wait for tiles to be sent
            images = yield

            # Wait for the last untiling thread to finish
            if self._tile_thread != None:
                self._tile_thread.result()

            # start a thread to untile the data
            b = bn + images.shape[0]
            self._tile_thread = thread_pool.submit(self._assemble_tiles, images, X[bn:b], Y[bn:b], Z[bn:b], C[bn:b],
                                                    T[bn:b])
            bn = b

            # Save a supertile if a thread is available
            if self._raw_buffer.qsize() > 0:
                if self._put_thread != None:
                    self._put_thread.result()
                self._put_thread = thread_pool.submit(self._put)

        # Wait for the final untiling thread to finish
        self._tile_thread.result()

        # Put the remaining pixels in the buffer into the _raw_buffer
        self._raw_buffer.put(self._pixel_buffer[:, 0:self.num_x() - self._tile_x_offset])

        # Save the last supertile
        if self._put_thread != None:
            self._put_thread.result()  # wait for the previous thread to finish
        self._put()  # no need to use a thread for final save

        thread_pool.shutdown()

        yield
